__all__ = ('ComponentSelect',)

import reprlib

from scarletio import copy_docs, include

from .component_base import ComponentBase
from .component_select_option import ComponentSelectOption
from .debug import (
    _debug_component_custom_id, _debug_component_enabled, _debug_component_max_values, _debug_component_min_values,
    _debug_component_options, _debug_component_placeholder
)
from .preinstanced import ComponentType


create_auto_custom_id = include('create_auto_custom_id')


class ComponentSelect(ComponentBase):
    """
    Select component.
    
    Attributes
    ----------
    custom_id : `str`
        Custom identifier to detect which component was used by the user.
    enabled : `bool`
        Whether the component is enabled.
    options : `None`, `tuple` of ``ComponentSelectOption``
        Options of the select.
    placeholder : `str`
        Placeholder text of the select.
    max_values : `int
        The maximal amount of options to select. Can be in range [1:25]. Defaults to `1`.
    min_values : `int`
        The minimal amount of options to select. Can be in range [1:15]. Defaults to `1`.
    
    Class Attributes
    ----------------
    type : ``ComponentType`` = `ComponentType.select`
        The component's type.
    """
    type = ComponentType.select
    
    __slots__ = ('custom_id', 'enabled', 'options', 'placeholder', 'max_values', 'min_values', )
    
    def __new__(cls, options, custom_id=None, *, enabled=True, placeholder=None, max_values=1, min_values=1):
        """
        Creates a new ``ComponentSelect`` with the given parameters.
        
        Parameters
        ----------
        options : `None` or (`list`, `tuple`) of ``ComponentSelectOption``
            Options of the select.
        custom_id : `None`, `str` = `None`, Optional
            Custom identifier to detect which component was used by the user.
        enabled : `bool` = `True`, Optional (Keyword only)
            Whether the button is enabled. Defaults to `True`.
        placeholder : `None`, `str` = `None`, Optional (Keyword only)
            Placeholder text of the select.
        max_values : `int` = `1`, Optional (Keyword only)
            The maximal amount of options to select. Can be in range [1:25].
        min_values : `int` = `1`, Optional (Keyword only)
            The minimal amount of options to select. Can be in range [1:15].
        
        Raises
        ------
        AssertionError
            - If `custom_id` is not given as `None`, `str`.
            - If `custom_id`'s length is out of range [0:100].
            - If `options` length is out from the expected range [1:25].
            - If `options` is neither `None` or (`list`, `tuple`) of ``ComponentSelectOption`` elements.
            - If `min_values` is not `int`.
            - If `min_values` is out of range [1:15].
            - If `max_values` is not `int`.
            - If `max_values` is out of range [1:25].
            - If `enabled` was not given as `bool`.
        """
        if __debug__:
            _debug_component_custom_id(custom_id)
            _debug_component_enabled(enabled)
            _debug_component_options(options)
            _debug_component_placeholder(placeholder)
            _debug_component_min_values(min_values)
            _debug_component_max_values(max_values)
        
        # custom_id
        if (custom_id is None) or (not custom_id):
            custom_id = create_auto_custom_id()
        
        # enabled
        # No additional checks
        
        # options
        if (options is not None):
            options = tuple(options)
            if (not options):
                options = None
        
        # placeholder
        if (placeholder is not None) and (not placeholder):
            placeholder = None
        
        # max_values
        # No additional checks
        
        # min_values
        # No additional checks
        
        self = object.__new__(cls)
        self.custom_id = custom_id
        self.enabled = enabled
        self.options = options
        self.placeholder = placeholder
        self.max_values = max_values
        self.min_values = min_values
        return self
    
    
    @classmethod
    @copy_docs(ComponentBase.from_data)
    def from_data(cls, data):
        self = object.__new__(cls)
        
        # custom_id
        self.custom_id = data['custom_id']
        
        # enabled
        self.enabled = not data.get('disabled', False)
        
        # options
        option_datas = data['options']
        if option_datas:
            options = tuple(ComponentSelectOption.from_data(option_data) for option_data in option_datas)
        else:
            options = None
        self.options = options
        
        # placeholder
        placeholder = data.get('placeholder', None)
        if (placeholder is not None) and (not placeholder):
            placeholder = None
        self.placeholder = placeholder
        
        # max_values
        self.max_values = data.get('max_values', 1)
        
        # min_values
        self.min_values = data.get('min_values', 1)
        
        return self
    
    
    @copy_docs(ComponentBase.to_data)
    def to_data(self):
        # type & custom_id
        data = {
            'type': self.type.value,
            'custom_id': self.custom_id,
        }
        
        # enabled
        if (not self.enabled):
            data['disabled'] = True
        
        # options
        options = self.options
        if options is None:
            options_value = []
        else:
            options_value = [option.to_data() for option in options]
        data['options'] = options_value
        
        # placeholder
        placeholder = self.placeholder
        if (placeholder is not None):
            data['placeholder'] = placeholder
        
        # max_values
        max_values = self.max_values
        if max_values != 1:
            data['max_values'] = max_values
        
        # min_values
        min_values = self.min_values
        if min_values != 1:
            data['min_values'] = min_values
        
        return data
    
    
    @copy_docs(ComponentBase.__repr__)
    def __repr__(self):
        repr_parts = ['<', self.__class__.__name__]
        
        # Descriptive fields : type
        
        # type
        type_ = self.type
        repr_parts.append(' type=')
        repr_parts.append(type_.name)
        repr_parts.append(' (')
        repr_parts.append(repr(type_.value))
        repr_parts.append(')')
        
        # System fields : custom_id & options
        
        # custom_id
        repr_parts.append(', custom_id=')
        repr_parts.append(reprlib.repr(self.custom_id))
        
        # options
        repr_parts.append(', options=')
        options = self.options
        if (options is None):
            repr_parts.append('[]')
        else:
            repr_parts.append('[')
            
            index = 0
            limit = len(options)
            
            while True:
                option = options[index]
                index += 1
                
                repr_parts.append(repr(option))
                
                if index == limit:
                    break
                
                repr_parts.append(', ')
                continue
            
            repr_parts.append(']')
        
        repr_parts.append('>')
        
        # Text fields : placeholder
        
        # placeholder
        placeholder = self.placeholder
        if (placeholder is not None):
            repr_parts.append(', placeholder=')
            repr_parts.append(repr(placeholder))
        
        # Optional descriptive fields: min_values & max_values & enabled
        
        # min_values
        min_values = self.min_values
        if min_values != 1:
            repr_parts.append(', min_values=')
            repr_parts.append(repr(min_values))
        
        # max_values
        max_values = self.max_values
        if max_values != 1:
            repr_parts.append(', max_values=')
            repr_parts.append(repr(max_values))
        
        # enabled
        enabled = self.enabled
        if (not enabled):
            repr_parts.append(', enabled=')
            repr_parts.append(repr(enabled))
        
        return ''.join(repr_parts)
    
    
    @copy_docs(ComponentBase.copy)
    def copy(self):
        new = object.__new__(type(self))
        
        # custom_id
        new.custom_id = self.custom_id
        
        # enabled
        new.enabled = self.enabled
        
        # options
        options = self.options
        if (options is not None):
            options = tuple(option.copy() for option in options)
        new.options = options
        
        # placeholder
        new.placeholder = self.placeholder
        
        # max_values
        new.max_values = self.max_values
        
        # min_values
        new.min_values = self.min_values
        
        return new
    
    
    def copy_with(self, **kwargs):
        """
        Copies the component and modifies the created one with the given parameters.
        
        Parameters
        ----------
        **kwargs : Keyword parameters
            Keyword parameters referencing attributes.
        
        Other Parameters
        ----------------
        custom_id : `None`, `str`, Optional (Keyword only)
            Custom identifier to detect which component was used by the user.
        
        enabled : `bool`, Optional (Keyword only)
            Whether the button is enabled. Defaults to `True`.
        
        options : `None` or (`list`, `tuple`) of ``ComponentSelectOption``, Optional (Keyword only)
            Options of the select.
        
        placeholder : `str`, Optional (Keyword only)
            Placeholder text of the select.
        
        max_values : `int`, Optional (Keyword only)
            The maximal amount of options to select. Can be in range [1:25]. Defaults to `1`.
        
        min_values : `int`, Optional (Keyword only)
            The minimal amount of options to select. Can be in range [1:15]. Defaults to `1`.
        
        Returns
        -------
        new : ``ComponentSelect``
        """
        # custom_id
        try:
            custom_id = kwargs.pop('custom_id')
        except KeyError:
            custom_id = self.custom_id
        else:
            if __debug__:
                _debug_component_custom_id(custom_id)
            
            if custom_id is None:
                custom_id = self.custom_id
        
        # enabled
        try:
            enabled = kwargs.pop('enabled')
        except KeyError:
            enabled = self.enabled
        else:
            if __debug__:
                _debug_component_enabled(enabled)
        
        # options
        try:
            options = kwargs.pop('options')
        except KeyError:
            options = self.options
            if (options is not None):
                options = tuple(option.copy() for option in options)
        else:
            if __debug__:
                _debug_component_options(options)
            
            if (options is not None):
                options = tuple(options)
                if (not options):
                    options = None
        
        # placeholder
        try:
            placeholder = kwargs.pop('placeholder')
        except KeyError:
            placeholder = self.placeholder
        else:
            if __debug__:
                _debug_component_placeholder(placeholder)
            
            if (placeholder is not None) and (not placeholder):
                placeholder = None
        
        # max_values
        try:
            max_values = kwargs.pop('max_values')
        except KeyError:
            max_values = self.max_values
        else:
            if __debug__:
                _debug_component_max_values(max_values)
        
        # min_values
        try:
            min_values = kwargs.pop('min_values')
        except KeyError:
            min_values = self.min_values
        else:
            if __debug__:
                _debug_component_min_values(min_values)
        
        if kwargs:
            raise TypeError(f'Unused or unsettable attributes: {kwargs!r}.')
        
        new = object.__new__(type(self))
        new.custom_id = custom_id
        new.enabled = enabled
        new.options = options
        new.placeholder = placeholder
        new.max_values = max_values
        new.min_values = min_values
        return new
    
    
    @copy_docs(ComponentBase.__eq__)
    def __eq__(self, other):
        if type(other) is not type(self):
            return NotImplemented
        
        # custom_id
        if self.custom_id != other.custom_id:
            return False
        
        # enabled
        if self.enabled != other.enabled:
            return False
        
        # options
        if self.options != other.options:
            return False
        
        # placeholder
        if self.placeholder != other.placeholder:
            return False
        
        # max_values
        if self.max_values != other.max_values:
            return False
        
        # min_values
        if self.min_values != other.min_values:
            return False
        
        return True
    
    
    @copy_docs(ComponentBase.__hash__)
    def __hash__(self):
        hash_value = self.type.value
        
        # custom_id
        hash_value ^= hash(self.custom_id)
        
        # enabled
        if self.enabled:
            hash_value ^= 1 << 8
        
        # options
        options = self.options
        if (options is not None):
            hash_value ^= len(options) << 12
            for option in options:
                hash_value ^= hash(option)
        
        # placeholder
        placeholder = self.placeholder
        if (placeholder is not None):
            hash_value ^= hash(placeholder)
        
        # max_values
        max_values = self.max_values
        if (max_values != 1):
            hash_value ^= (max_values << 18)
        
        # min_values
        min_values = self.min_values
        if (min_values != 1):
            min_values ^= (min_values << 22)
        
        return hash_value
    
    
    @copy_docs(ComponentBase._iter_components)
    def _iter_components(self):
        yield self
        
        options = self.options
        if (options is not None):
            for option in options:
                yield from option._iter_components()
    
    
    @copy_docs(ComponentBase._replace_direct_sub_components)
    def _replace_direct_sub_components(self, relation):
        options = self.options
        if (options is not None):
            self.options = tuple(relation.get(option, option) for option in options)
    
    
    @copy_docs(ComponentBase._iter_direct_sub_components)
    def _iter_direct_sub_components(self):
        options = self.options
        if (options is not None):
            yield from options



__all__ = ('bind',)

from collections import OrderedDict

from scarletio import RichAttributeErrorBaseType, WeakItemDictionary, WeakKeyDictionary, copy_docs


def bind(bind_to, bind_with, name, *, weak=False, weak_cache_size=0):
    """
    A cool hata features, that lets you bind object to existing one.
    
    Parameters
    ----------
    bind_to : `type`
        The type to bind to.
    bind_with : `type`
        The type to bind with.
    name : `str`
        The name of the binding.
    weak : `bool`, Optional (Keyword only) = `False`
        Whether double weak binding should be implemented.
    weak_cache_size : `int`, Optional (Keyword only) = `0`
        Whether double weak binding should have binding cache. Can be useful when binding requires accessing outer
        resources (like database).
    
    Raises
    ------
    TypeError
        - If `bind_to` is not a type.
        - If `bind_to`-s do not support weakreferencing.
        - If `bind_to` has already an attribute named `name`.
        - If `bind_with` is not weakreferable, but `weak` binding is created.
        - If `weak_cache_size` is not `int`.
    
    Examples
    --------
    Generic binder:
    ```py
    from hata import ClientUserBase, bind
    
    class Inventory:
        def __init__(self, parent_self):
            self.user_id = parent_self.id
            self.inventory = []
        
        def add_item(self, name):
            self.inventory.append(name)
        
        def give_user(self, item, recipient):
            self.inventory.remove(item)
            recipient.inventory.add_item(item)
    
    bind(ClientUserBase, Inventory, 'inventory')
    
    # Usage:
    user.inventory.add_item('cake')
    ```
    
    Descriptor binder:
    ```py
    from hata import ClientUserBase, bind
    
    class ValueBinder:
        def __init__(self, parent_self):
            self.value = 10
        
        def __get__(self, instance, type_):
            return self.value
        
        def __set__(self, instance, value):
            self.value = value
    
    bind(ClientUserBase, ValueBinder, 'value')
    
    # Usage:
    user.inventory_size += 10
    ```
    Descriptor binders are familiar ot normal descriptors, except, they are created per object and not per class.
    
    Sometimes you want to clear extra binders with default values. For this you might want to define the `__bool__`
    method.
    ```py
    DEFAULT_VALUE = 10
    
    class ValueBinder:
        def __init__(self, parent_self):
            self.value = DEFAULT_VALUE
        
        def __get__(self, instance, type_):
            return self.value
        
        def __set__(self, instance, value):
            self.value = value
        
        def __bool__(self):
            return (self.value != DEFAULT_VALUE)
    
    
    bind(ClientUserBase, ValueBinder, 'value')
    
    # Usage
    ClientUserBase.value.clear()
    ```
    
    A more advanced concept is reloading modules, which use bindings. For this define `__getstate__` and the
    `__setstate__` magic methods.
    
    > Also defining `__bool__` is not required, but recommended.
    ```py
    class Inventory:
        def __init__(self, parent_self):
            self.user_id = parent_self.id
            self.inventory = []
        
        def add_item(self, name):
            self.inventory.append(name)
        
        def give_user(self, item, recipient):
            self.inventory.remove(item)
            recipient.inventory.add_item(item)
        
        def __getstate__(self):
            return (self.user_id, self.inventory)
        
        def __setstate(self, state):
            self.user_id, self.inventory = state
    
    bind(ClientUserBase, Inventory, 'inventory')
    ```
    """
    if not isinstance(bind_to, type):
        raise TypeError(
            f'`bind_to` can be `type`, got {bind_to.__class__.__name__}; {bind_to!r}.'
        )
    
    name_space = bind_to.__dict__
    if ('__weakref__' in name_space) or ('__slots__' not in name_space):
        raise TypeError(
            f'`bind_to`-s must support weakreferencing, got {bind_to!r}.'
        )
    
    
    if weak:
        binded_name_space = bind_with.__dict__
        if ('__weakref__' in binded_name_space) or ('__slots__' not in binded_name_space):
            raise TypeError(
                f'`bind_with`-s must support weakreferencing for weak binding, got {bind_with!r}.'
            )
    
    if hasattr(bind_to, name):
        old_binder = getattr(bind_to, name)
        if not isinstance(old_binder, ObjectBinderBase):
            raise TypeError(
                f'`bind_to` already has an attribute named, as bind_to={bind_to!r}, name={name!r}.'
            )
    
    else:
        old_binder = None
    
    if not isinstance(weak_cache_size, int):
        raise TypeError(
            f'`weak_cache_size` can be `int`, got {weak_cache_size.__class__.__name__}; {weak_cache_size!r}.'
        )
    
    is_descriptor = getattr(bind_with, '__get__', None) is not None
    if weak:
        if is_descriptor:
            new_binder = DescriptorObjectWeakBinder(name, bind_with, weak_cache_size)
        else:
            new_binder = GenericObjectWeakBinder(name, bind_with, weak_cache_size)
    else:
        if is_descriptor:
            new_binder = DescriptorObjectBinder(name, bind_with)
        else:
            new_binder = GenericObjectBinder(name, bind_with)
    
    if (
        (old_binder is not None) and
        old_binder.supports_state_transfer and
        new_binder.supports_state_transfer and
        (old_binder.get_type_key() == new_binder.get_type_key())
    ):
        new_binder.set_states(old_binder.get_states(bind_to))
    
    setattr(bind_to, name, new_binder)


class ObjectBinderBase(RichAttributeErrorBaseType):
    """
    Object binder.
    
    Attributes
    ----------
    name : `str`
        The field's name.
    relations : ``WeakKeyDictionary``
        The binding relation mapping.
    supports_clearing : `bool`
        Whether clearing the bound objects is supported.
    supports_state_transfer : `bool`
        Whether the bound objects support state transfer.
    type : `type`
        The binded class.
    
    Class Attributes
    ----------------
    RELATIONSHIP_CONTAINER_TYPE : `type` = ``WeakKeyDictionary``
        The type which holds the relations.
    """
    __slots__ = ('name', 'relations', 'supports_clearing', 'supports_state_transfer', 'type')
    
    RELATIONSHIP_CONTAINER_TYPE = WeakKeyDictionary
    
    def __new__(cls, name, type_):
        """
        Creates a new object binder instance.
        
        Parameters
        ----------
        name : `str`
            The field's name.
        type_ : `type`
            The binded class.
        """
        bool_method = getattr(type_, '__bool__', None)
        if (bool_method is None):
            supports_clearing = False
        else:
            supports_clearing = True
        
        get_state_method = getattr(type_, '__getstate__', None)
        set_state_method = getattr(type_, '__setstate__', None)
        if (get_state_method is None) or (set_state_method is None):
            supports_state_transfer = False
        else:
            supports_state_transfer = True
        
        self = object.__new__(cls)
        self.relations = cls.RELATIONSHIP_CONTAINER_TYPE()
        self.name = name
        self.type = type_
        
        self.supports_clearing = supports_clearing
        self.supports_state_transfer = supports_state_transfer
        
        return self
    
    
    def __get__(self, instance, type_):
        """
        Gets the field's value.
        
        Parameters
        ----------
        instance : `Any`
            The instance itself.
        type_ : `type`
            The instance's type.
        
        Returns
        -------
        value : `self` / `Any`
            Returns itself if called from class.
        """
        if (instance is None):
            return self
    
    
    def __set__(self, instance, value):
        """
        Sets the value to the instance.
        
        Parameters
        ----------
        instance : `Any`
            The parent instance to set the value to.
        value : `Any`
            The value to set.
        
        Raises
        ------
        AttributeError
            Read only attribute.
        """
        raise AttributeError(f'Read only attribute: {self.name!r}.')
    
    
    def __delete__(self, instance):
        """
        Deletes the value of the instance.
        
        Parameters
        ----------
        instance : `Any`
            The parent instance to delete the attribute of.
        
        Raises
        ------
        AttributeError
            Cannot delete attribute.
        """
        raise AttributeError(f'Cannot delete attribute: {self.name!r}.')
    
    
    def __repr__(self):
        """Returns the object binder representation."""
        repr_parts = ['<', self.__class__.__name__]
        
        repr_parts.append(' name=')
        repr_parts.append(repr(self.name))
        
        repr_parts.append(', type=')
        repr_parts.append(repr(self.type))
        
        length = len(self.relations)
        if length:
            repr_parts.append(', length=')
            repr_parts.append(repr(length))
        
        supports_clearing = self.supports_clearing
        if supports_clearing:
            repr_parts.append(', supports_clearing=')
            repr_parts.append(repr(supports_clearing))
        
        supports_state_transfer = self.supports_state_transfer
        if supports_state_transfer:
            repr_parts.append(', supports_state_transfer=')
            repr_parts.append(repr(supports_state_transfer))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def clear(self):
        """
        Clears the bound objects.
        
        Only those objects are cleared, which return `False` from their `__bool__` method.
        
        Returns
        -------
        cleared : `int`
            The amount of objects freed.
        """
        if not self.supports_clearing:
            return 0
        
        relations = self.relations
        to_clear = []
        
        for key, value in relations.items():
            if not value:
                to_clear.append(key)
        
        for key in to_clear:
            del relations[key]
        
        return len(to_clear)
    
    
    def get_type_key(self):
        """
        Returns the type key of the object binder.
        
        It is used when transferring down states to match types.
        
        Returns
        -------
        type_key : `tuple` ((`None`, `str`), (`None`, `str`))
        """
        type_ = self.type
        return getattr(type_, '__module__', None), getattr(type_, '__name__', None)
    
    
    def iter_true_items(self):
        """
        Iterates over all non empty bound objects & keys.
        
        This method is an iterable generator.
        
        Yields
        ------
        key : `Any`
            Objects to bound to.
        value : ``.type``
            The bound object.
        """
        if self.supports_clearing:
            for key, value in self.relations.items():
                if value:
                    yield key, value
        
        else:
            yield from self.relations.items()
    
    
    def get_states(self, type_):
        """
        Gets the given object's states if applicable for the given `type_` instances.
        
        Returns
        -------
        states : `list` of `tuple` (`Any`, `Any`)
        """
        states = []
        if self.supports_state_transfer:
            get_state = self.type.__getstate__
            
            for key, value in self.iter_true_items():
                if not isinstance(key, type_):
                    continue
                
                state = get_state(value)
                states.append((key, state))
        
        return states
    
    
    def set_states(self, states):
        """
        Sets states of an another binder to this one.
        
        Parameters
        ----------
        states : `list` of `tuple` (`Any`, `Any`)
        """
        if self.supports_state_transfer:
            type_ = self.type
            set_state = type_.__setstate__
            relations = self.relations
            
            for key, state in states:
                value = object.__new__(type_)
                set_state(value, state)
                
                relations[key] = value


@copy_docs(ObjectBinderBase)
class GenericObjectBinder(ObjectBinderBase):
    __slots__ = ()
    
    @copy_docs(ObjectBinderBase.__get__)
    def __get__(self, instance, type_):
        if instance is None:
            return self
        
        try:
            bind_object = self.relations[instance]
        except KeyError:
            bind_object = self.type(instance)
            self.relations[instance] = bind_object
        
        return bind_object


class DescriptorObjectBinder(ObjectBinderBase):
    """
    Object binder for descriptors-likes.
    
    Attributes
    ----------
    relations : ``WeakKeyDictionary``
        The binding relation mapping.
    name : `str`
        The field's name.
    type : `type`
        The binded class.
    fdel : `None`, `Any`
        Deleter function if applicable.
    fget : `None`, `Any`
        Getter function if applicable.
    fset : `None`, `Any`
        Setter function if applicable.
    
    Class Attributes
    ----------------
    RELATIONSHIP_CONTAINER_TYPE : `type` = ``WeakKeyDictionary``
        The type which holds the relations.
    """
    __slots__ = ('fdel', 'fget', 'fset')
    
    @copy_docs(ObjectBinderBase.__new__)
    def __new__(cls, name, type_):
        self = ObjectBinderBase.__new__(cls, name, type_)
        self.fdel = getattr(type_, '__delete__', None)
        self.fget = getattr(type_, '__get__')
        self.fset = getattr(type_, '__set__', None)
        return self
    
    
    @copy_docs(ObjectBinderBase.__get__)
    def __get__(self, instance, type_):
        if instance is None:
            return self
        
        try:
            bind_object = self.relations[instance]
        except KeyError:
            bind_object = self.type(instance)
            self.relations[instance] = bind_object
        
        return self.fget(bind_object, instance, type_)
    
    
    @copy_docs(ObjectBinderBase.__get__)
    def __set__(self, instance, value):
        fset = self.fset
        if fset is None:
            raise AttributeError(f'Read only attribute: {self.name!r}.')
        
        try:
            bind_object = self.relations[instance]
        except KeyError:
            bind_object = self.type(instance)
            self.relations[instance] = bind_object
        
        return fset(bind_object, instance, value)
    
    
    @copy_docs(ObjectBinderBase.__delete__)
    def __delete__(self, instance):
        fdel = self.fdel
        if fdel is None:
            raise AttributeError(f'Cannot delete attribute: {self.name!r}.')
        
        try:
            bind_object = self.relations[instance]
        except KeyError:
            bind_object = self.type(instance)
            self.relations[instance] = bind_object
        
        return fdel(bind_object, instance)



def _cache_bind(cache, cache_size, bind_object):
    """
    Caches the given bind object if applicable.
    
    Parameters
    ----------
    cache : `None`, `OrderedDict`
        Cache if applicable.
    cache_size : `int`
        The maximal size of the cache.
    bind_object : `Any`
        The bind object to cache.
    """
    if (cache is not None):
        if bind_object in cache:
            cache.move_to_end(bind_object)
        
        else:
            cache[bind_object] = bind_object
            
            if len(cache) >= cache_size:
                del cache[next(iter(cache))]


class GenericObjectWeakBinder(GenericObjectBinder):
    """
    Object binder.
    
    Attributes
    ----------
    name : `str`
        The field's name.
    relations : ``WeakKeyDictionary``
        The binding relation mapping.
    supports_clearing : `bool`
        Whether clearing the bound objects is supported.
    supports_state_transfer : `bool`
        Whether the bound objects support state transfer.
    type : `type`
        The binded class.
    
    cache : `None`, ``OrderedDict``
        Cache dictionary, to avoid repeated initializing.
    cache_size : `int`
        The maximal size of the cache.
    
    Class Attributes
    ----------------
    RELATIONSHIP_CONTAINER_TYPE : `type` = ``WeakItemDictionary``
        The type which holds the relations.
    """
    __slots__ = ('cache', 'cache_size',)
    
    RELATIONSHIP_CONTAINER_TYPE = WeakItemDictionary
    
    def __new__(cls, name, type_, cache_size):
        """
        Creates a new object binder instance.
        
        Parameters
        ----------
        name : `str`
            The field's name.
        type_ : `type`
            The binded class.
        cache_size : `int`
            The maximal size of the cache.
        """
        if cache_size > 0:
            cache = OrderedDict()
        else:
            cache = None
        
        self = GenericObjectBinder.__new__(cls, name, type_)
        
        self.cache = cache
        self.cache_size = cache_size
        
        return self
    
    
    @copy_docs(GenericObjectBinder.__get__)
    def __get__(self, instance, type_):
        bind_object = GenericObjectBinder.__get__(self, instance, type_)
        if (bind_object is not self):
            _cache_bind(self.cache, self.cache_size, bind_object)
        
        return bind_object


class DescriptorObjectWeakBinder(DescriptorObjectBinder):
    """
    Object binder for descriptors-likes.
    
    Attributes
    ----------
    relations : ``WeakKeyDictionary``
        The binding relation mapping.
    name : `str`
        The field's name.
    type : `type`
        The binded class.
    fdel : `None`, `Any`
        Deleter function if applicable.
    fget : `None`, `Any`
        Getter function if applicable.
    fset : `None`, `Any`
        Setter function if applicable.
    
    cache : `None`, ``OrderedDict``
        Cache dictionary, to avoid repeated initializing.
    cache_size : `int`
        The maximal size of the cache.
    
    Class Attributes
    ----------------
    RELATIONSHIP_CONTAINER_TYPE : `type` = ``WeakItemDictionary``
        The type which holds the relations.
    """
    __slots__ = GenericObjectWeakBinder.__slots__
    
    RELATIONSHIP_CONTAINER_TYPE = GenericObjectWeakBinder.RELATIONSHIP_CONTAINER_TYPE
    
    @copy_docs(GenericObjectWeakBinder.__new__)
    def __new__(cls, name, type_, cache_size):
        if cache_size > 0:
            cache = OrderedDict()
        else:
            cache = None
        
        self = DescriptorObjectBinder.__new__(cls, name, type_)
        
        self.cache = cache
        self.cache_size = cache_size
        
        return self
    
    
    @copy_docs(DescriptorObjectBinder.__get__)
    def __get__(self, instance, type_):
        bind_object = DescriptorObjectBinder.__get__(self, instance, type_)
        if (bind_object is not self):
            _cache_bind(self.cache, self.cache_size, bind_object)
        
        return bind_object



__all__ = (
    'ChannelMix', 'Distortion', 'Equalizer', 'Filter', 'Karaoke', 'LowPass', 'Rotation', 'Timescale', 'Tremolo',
    'Vibrato', 'Volume'
)

from scarletio import RichAttributeErrorBaseType, copy_docs

from .constants import (
    LAVALINK_BAND_COUNT, LAVALINK_KEY_FILTER_CHANNEL_MIX, LAVALINK_KEY_FILTER_CHANNEL_MIX_LEFT_TO_LEFT,
    LAVALINK_KEY_FILTER_CHANNEL_MIX_LEFT_TO_RIGHT, LAVALINK_KEY_FILTER_CHANNEL_MIX_RIGHT_TO_LEFT,
    LAVALINK_KEY_FILTER_CHANNEL_MIX_RIGHT_TO_RIGHT, LAVALINK_KEY_FILTER_DISTORTION,
    LAVALINK_KEY_FILTER_DISTORTION_COS_OFFSET, LAVALINK_KEY_FILTER_DISTORTION_COS_SCALE,
    LAVALINK_KEY_FILTER_DISTORTION_OFFSET, LAVALINK_KEY_FILTER_DISTORTION_SCALE,
    LAVALINK_KEY_FILTER_DISTORTION_SIN_OFFSET, LAVALINK_KEY_FILTER_DISTORTION_SIN_SCALE,
    LAVALINK_KEY_FILTER_DISTORTION_TAN_OFFSET, LAVALINK_KEY_FILTER_DISTORTION_TAN_SCALE, LAVALINK_KEY_FILTER_EQUALIZER,
    LAVALINK_KEY_FILTER_EQUALIZER_BAND, LAVALINK_KEY_FILTER_EQUALIZER_GAIN, LAVALINK_KEY_FILTER_KARAOKE,
    LAVALINK_KEY_FILTER_KARAOKE_FILTER_BAND, LAVALINK_KEY_FILTER_KARAOKE_FILTER_WIDTH,
    LAVALINK_KEY_FILTER_KARAOKE_LEVEL, LAVALINK_KEY_FILTER_KARAOKE_MONO_LEVEL, LAVALINK_KEY_FILTER_LOW_PASS,
    LAVALINK_KEY_FILTER_LOW_PASS_SMOOTHING, LAVALINK_KEY_FILTER_ROTATION, LAVALINK_KEY_FILTER_ROTATION_ROTATION,
    LAVALINK_KEY_FILTER_TIMESCALE, LAVALINK_KEY_FILTER_TIMESCALE_PITCH, LAVALINK_KEY_FILTER_TIMESCALE_RATE,
    LAVALINK_KEY_FILTER_TIMESCALE_SPEED, LAVALINK_KEY_FILTER_TREMOLO, LAVALINK_KEY_FILTER_TREMOLO_DEPTH,
    LAVALINK_KEY_FILTER_TREMOLO_FREQUENCY, LAVALINK_KEY_FILTER_VIBRATO, LAVALINK_KEY_FILTER_VIBRATO_DEPTH,
    LAVALINK_KEY_FILTER_VIBRATO_FREQUENCY, LAVALINK_KEY_FILTER_VOLUME
)

FILTER_IDENTIFIER_NONE = 0
FILTER_IDENTIFIER_EQUALIZER = 1
FILTER_IDENTIFIER_KARAOKE = 2
FILTER_IDENTIFIER_TIMESCALE = 3
FILTER_IDENTIFIER_TREMOLO = 4
FILTER_IDENTIFIER_VIBRATO = 5
FILTER_IDENTIFIER_ROTATION = 6
FILTER_IDENTIFIER_DISTORTION = 7
FILTER_IDENTIFIER_CHANNEL_MIX = 8
FILTER_IDENTIFIER_LOW_PASS = 9
FILTER_IDENTIFIER_VOLUME = 10

# Use these as references:
#
# https://github.com/freyacodes/Lavalink/blob/master/LavalinkServer/src/main/java/lavalink/server/player/filters/filterConfigs.kt
# https://github.com/freyacodes/Lavalink/blob/master/IMPLEMENTATION.md#using-filters
# https://github.com/natanbc/lavadsp/tree/master/src/main/java/com/github/natanbc/lavadsp

class Filter(RichAttributeErrorBaseType):
    """
    Represents filters applied to a solar client.
    
    Adding a filter can have adverse effects on performance. Filters force the lava player to decode all audio to PCM,
    even if the input was already in the Opus format that Discord uses. This means decoding and encoding audio that
    would normally require very little processing. This is often the case with YouTube videos.
    
    Class Attributes
    ----------------
    identifier : `int`
        The filter type's internal identifier.
    json_key : `str`
        The key of the filter used when serializing it.
    """
    identifier = FILTER_IDENTIFIER_NONE
    json_key = ''
    
    __slots__ = ()
    
    def __new__(cls):
        """
        Creates a new filter instance.
        """
        return object.__new__(cls)
    
    
    def __eq__(self, other):
        """Returns whether the two filters are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        return True
    
    
    def __hash__(self):
        """Returns the filter's hash value."""
        return 0
    
    
    def __repr__(self):
        """Returns the filter's representation."""
        return f'<{self.__class__.__name__}>'
    
    
    def __bool__(self):
        """Returns whether the filter was changed."""
        return False
    
    
    def to_data(self):
        """
        Converts the filter to a json serializable object.
        
        Returns
        -------
        data : `dict` of (`str`, `Any`) items
        """
        return {}



class ChannelMix(Filter):
    """
    Mixes both channels (left and right), with a configurable factor on how much each channel affects the other.
    
    With the defaults, both channels are kept independent from each other.
    
    Setting all factors to 0.5 means both channels get the same audio.
    
    Attributes
    ----------
    _left_to_left : `float`
        Left-to-left factor.
    _left_to_right : `float`
        Left-to-right factor.
    _right_to_left : `float`
        Right-to-left factor.
    _right_to_right : `float`
        Right-to-right factor.
    
    Class Attributes
    ----------------
    identifier : `int`
        The filter type's internal identifier.
    json_key : `str`
        The key of the filter used when serializing it.
    """
    identifier = FILTER_IDENTIFIER_CHANNEL_MIX
    json_key = LAVALINK_KEY_FILTER_CHANNEL_MIX
    
    __slots__ = ('_left_to_left', '_left_to_right', '_right_to_left', '_right_to_right')
    
    def __new__(cls, left_to_left, left_to_right, right_to_right, right_to_left):
        """
        Creates a new channel mix filter.
        
        Parameters
        ----------
        left_to_left : `float`
            Left-to-left factor.
        left_to_right : `float`
            Left-to-right factor.
        right_to_right : `float`
            Right-to-right factor.
        right_to_left : `float`
            Right-to-left factor.
        
        Raises
        ------
        TypeError
            - If `left_to_left` is not `float`.
            - If `left_to_right` is not `float`.
            - If `right_to_right` is not `float`.
            - If `right_to_left` is not `float`.
        """
        if not isinstance(left_to_left, float):
            raise TypeError(
                f'`left_to_left` can be `float`, got {left_to_left.__class__.__name__}; {left_to_left!r}.'
            )
        
        if not isinstance(left_to_right, float):
            raise TypeError(
                f'`left_to_right` can be `float`, got {left_to_right.__class__.__name__}; {left_to_right!r}.'
            )
        
        if not isinstance(right_to_right, float):
            raise TypeError(
                f'`right_to_right` can be `float`, got {right_to_right.__class__.__name__}; {right_to_right!r}.'
            )
        
        if not isinstance(right_to_left, float):
            raise TypeError(
                f'`right_to_left` can be `float`, got {right_to_left.__class__.__name__}; {right_to_left!r}.'
            )
        
        
        self = object.__new__(cls)
        self._left_to_left = left_to_left
        self._left_to_right = left_to_right
        self._right_to_left = right_to_left
        self._right_to_right = right_to_right
        return self
    
    
    @copy_docs(Filter.__eq__)
    def __eq__(self, other):
        if type(self) is not type(other):
            return NotImplemented
        
        if self._left_to_left != other._left_to_left:
            return False
        
        if self._left_to_right != other._left_to_right:
            return False
        
        if self._right_to_left != other._right_to_left:
            return False
        
        if self._right_to_right != other._right_to_right:
            return False
        
        return True
    
    
    @copy_docs(Filter.__hash__)
    def __hash__(self):
        hash_value = 0
        
        left_to_left = self._left_to_left
        if (left_to_left != 1.0):
            hash_value ^= (1 << 4)
            hash_value ^= hash(left_to_left)
        
        left_to_right = self._left_to_right
        if (left_to_right != 0.0):
            hash_value ^= (1 << 8)
            hash_value ^= hash(left_to_right)
        
        right_to_left = self._right_to_left
        if (right_to_left != 0.0):
            hash_value ^= (1 << 12)
            hash_value ^= hash(right_to_left)
        
        right_to_right = self._right_to_right
        if (right_to_right != 0.0):
            hash_value ^= (1 << 16)
            hash_value ^= hash(right_to_right)
        
        return hash_value
    
    
    @copy_docs(Filter.__repr__)
    def __repr__(self):
        repr_parts = ['<', self.__class__.__name__]
        
        repr_parts.append(' left_to_left=')
        repr_parts.append(format(self._left_to_left, '.02f'))
        
        repr_parts.append(' left_to_right=')
        repr_parts.append(format(self._left_to_right, '.02f'))
        
        repr_parts.append(' right_to_right=')
        repr_parts.append(format(self._right_to_right, '.02f'))
        
        repr_parts.append(' right_to_left=')
        repr_parts.append(format(self._right_to_left, '.02f'))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    @copy_docs(Filter.__bool__)
    def __bool__(self):
        if self._left_to_left != 1.0:
            return True
        
        if self._left_to_right != 0.0:
            return True
        
        if self._right_to_right != 0.0:
            return True
        
        if self._right_to_left != 1.0:
            return True
        
        return False
    
    
    @copy_docs(Filter.to_data)
    def to_data(self):
        return {
            LAVALINK_KEY_FILTER_CHANNEL_MIX_LEFT_TO_LEFT: self._left_to_left,
            LAVALINK_KEY_FILTER_CHANNEL_MIX_LEFT_TO_RIGHT: self._left_to_right,
            LAVALINK_KEY_FILTER_CHANNEL_MIX_RIGHT_TO_LEFT: self._right_to_right,
            LAVALINK_KEY_FILTER_CHANNEL_MIX_RIGHT_TO_RIGHT: self._right_to_left,
        }


class Distortion(Filter):
    """
    Distortion effect.
    
    It can generate some pretty unique audio effects.
    
    Attributes
    ----------
    _cos_offset : `float`
        Cos offset.
    _cos_scale : `float`
        Cos scale.
    _offset : `float`
        Offset.
    _scale : `float`
        Scale.
    _sin_offset : `float`
        Sin offset.
    _sin_scale : `float`
        Sin scale.
    _tan_offset : `float`
        Tan offset.
    _tan_scale : `float`
        Tan scale.
    
    Class Attributes
    ----------------
    identifier : `int`
        The filter type's internal identifier.
    json_key : `str`
        The key of the filter used when serializing it.
    """
    identifier = FILTER_IDENTIFIER_DISTORTION
    json_key = LAVALINK_KEY_FILTER_DISTORTION
    
    __slots__ = (
        '_cos_offset', '_cos_scale', '_offset', '_scale', '_sin_offset', '_sin_scale', '_tan_offset', '_tan_scale'
    )
    
    def __new__(cls, *, sin_offset=0.0, sin_scale=1.0, cos_offset=0.0, cos_scale=1.0, tan_offset=0.0, tan_scale=1.0,
            offset=0.0, scale=1.0):
        """
        Creates a new distortion filter.
        
        Parameters
        ----------
        sin_offset : `float` = `0.0`, Optional (Keyword only)
            Sin offset.
        
        sin_scale : `float` = `1.0`, Optional (Keyword only)
            Sin scale.
        
        cos_offset : `float` = `0.0`, Optional (Keyword only)
            Cos offset.
        
        cos_scale : `float` = `1.0`, Optional (Keyword only)
            Cos scale.
        
        tan_offset : `float` = `0.0`, Optional (Keyword only)
            Tan offset.
        
        tan_scale : `float` = `1.0`, Optional (Keyword only)
            Tan scale.
            
        offset : `float` = `0.0`, Optional (Keyword only)
            Offset.
        
        scale : `float` = `1.0`, Optional (Keyword only)
            Scale.
        
        Raises
        ------
        TypeError
            - If `sin_offset` is not `float`.
            - If `sin_scale` is not `float`.
            - If `cos_offset` is not `float`.
            - If `cos_scale` is not `float`.
            - If `tan_offset` is not `float`.
            - If `tan_scale` is not `float`.
            - If `offset` is not `float`.
            - If `scale` is not `float`.
        """
        if not isinstance(sin_offset, float):
            raise TypeError(
                f'`sin_offset` can be `float`, got {sin_offset.__class__.__name__}; {sin_offset!r}.'
            )
        
        if not isinstance(sin_scale, float):
            raise TypeError(
                f'`sin_scale` can be `float`, got {sin_scale.__class__.__name__}; {sin_scale!r}.'
            )
        
        if not isinstance(cos_offset, float):
            raise TypeError(
                f'`cos_offset` can be `float`, got {cos_offset.__class__.__name__}; {cos_offset!r}.'
            )
        
        if not isinstance(cos_scale, float):
            raise TypeError(
                f'`cos_scale` can be `float`, got {cos_scale.__class__.__name__}; {cos_scale!r}.'
            )
        
        if not isinstance(tan_offset, float):
            raise TypeError(
                f'`tan_offset` can be `float`, got {tan_offset.__class__.__name__}; {tan_offset!r}.'
            )
        
        if not isinstance(tan_scale, float):
            raise TypeError(
                f'`tan_scale` can be `float`, got {tan_scale.__class__.__name__}; {tan_scale!r}.'
            )
        
        if not isinstance(offset, float):
            raise TypeError(
                f'`offset` can be `float`, got {offset.__class__.__name__}; {offset!r}.'
            )
        
        if not isinstance(scale, float):
            raise TypeError(
                f'`scale` can be `float`, got {scale.__class__.__name__}; {scale!r}.'
            )
        
        self = object.__new__(cls)
        self._cos_offset = cos_offset
        self._cos_scale = cos_scale
        self._offset = offset
        self._scale = scale
        self._sin_offset = sin_offset
        self._sin_scale = sin_scale
        self._tan_offset = tan_offset
        self._tan_scale = tan_scale
        return self
    
    
    @copy_docs(Filter.__eq__)
    def __eq__(self, other):
        if type(self) is not type(other):
            return NotImplemented
        
        if self._cos_offset != other._cos_offset:
            return False
        
        if self._cos_scale != other._cos_scale:
            return False
        
        if self._offset != other._offset:
            return False
        
        if self._scale != other._scale:
            return False
        
        if self._sin_offset != other._sin_offset:
            return False
        
        if self._sin_scale != other._sin_scale:
            return False
        
        if self._tan_offset != other._tan_offset:
            return False
        
        if self._tan_scale != other._tan_scale:
            return False
        
        return True
    
    
    @copy_docs(Filter.__hash__)
    def __hash__(self):
        hash_value = 0
        
        cos_offset = self._cos_offset
        if (cos_offset != 0.0):
            hash_value ^= hash(cos_offset)
        
        cos_scale = self._cos_scale
        if (cos_scale != 1.0):
            hash_value ^= hash(cos_scale)
        
        offset = self._offset
        if (offset != 0.0):
            hash_value ^= hash(offset)
        
        scale = self._scale
        if (scale != 1.0):
            hash_value ^= hash(scale)
        
        sin_offset = self._sin_offset
        if (sin_offset != 0.0):
            hash_value ^= hash(sin_offset)
        
        sin_scale = self._sin_scale
        if (sin_scale != 1.0):
            hash_value ^= hash(sin_scale)
        
        tan_offset = self._tan_offset
        if (tan_offset != 0.0):
            hash_value ^= hash(tan_offset)
        
        tan_scale = self._tan_scale
        if (tan_scale != 1.0):
            hash_value ^= hash(tan_scale)
        
        return hash_value
    
    
    @copy_docs(Filter.__repr__)
    def __repr__(self):
        repr_parts = ['<', self.__class__.__name__]
        
        sin_offset = self._sin_offset
        if (sin_offset != 0.0):
            field_added = True
        
            repr_parts.append(' sin_offset=')
            repr_parts.append(format(sin_offset, '.02f'))
        
        else:
            field_added = False
        
        
        sin_scale = self._sin_scale
        if (sin_scale != 1.0):
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' sin_scale=')
            repr_parts.append(format(sin_scale, '.02f'))
        
        
        cos_offset = self._cos_offset
        if (cos_offset != 0.0):
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' cos_offset=')
            repr_parts.append(format(cos_offset, '.02f'))

        
        cos_scale = self._cos_scale
        if (cos_scale != 1.0):
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' cos_scale=')
            repr_parts.append(format(cos_scale, '.02f'))
        
        
        tan_offset = self._tan_offset
        if (tan_offset != 0.0):
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' tan_offset=')
            repr_parts.append(format(tan_offset, '.02f'))

        
        tan_scale = self._tan_scale
        if (tan_scale != 1.0):
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' tan_scale=')
            repr_parts.append(format(tan_scale, '.02f'))
        
        
        offset = self._offset
        if (offset != 0.0):
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' offset=')
            repr_parts.append(format(offset, '.02f'))

        
        scale = self._scale
        if (scale != 1.0):
            if field_added:
                repr_parts.append(',')
            
            repr_parts.append(' scale=')
            repr_parts.append(format(scale, '.02f'))
        
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    @copy_docs(Filter.__bool__)
    def __bool__(self):
        if self._cos_offset != 0.0:
            return True
        
        if self._cos_scale != 1.0:
            return True
        
        if self._sin_offset == 0.0:
            return True
        
        if self._sin_scale != 1.0:
            return True
        
        if self._tan_offset == 0.0:
            return True
        
        if self._tan_scale != 1.0:
            return True
        
        if self._offset == 0.0:
            return True
        
        if self._scale != 1.0:
            return True
        
        return False
    
    
    @copy_docs(Filter.to_data)
    def to_data(self):
        return {
            LAVALINK_KEY_FILTER_DISTORTION_COS_OFFSET: self._cos_offset,
            LAVALINK_KEY_FILTER_DISTORTION_COS_SCALE: self._cos_scale,
            LAVALINK_KEY_FILTER_DISTORTION_SIN_OFFSET: self._sin_offset,
            LAVALINK_KEY_FILTER_DISTORTION_SIN_SCALE: self._sin_scale,
            LAVALINK_KEY_FILTER_DISTORTION_TAN_OFFSET: self._tan_offset,
            LAVALINK_KEY_FILTER_DISTORTION_TAN_SCALE: self._tan_scale,
            LAVALINK_KEY_FILTER_DISTORTION_OFFSET: self._offset,
            LAVALINK_KEY_FILTER_DISTORTION_SCALE: self._scale,
        }


class Equalizer(Filter):
    """
    There are 15 bands (0-14) that can be changed.
    
    "gain" is the multiplier for the given band. The default value is `0`. Valid values range from `-0.25` to `1.0`,
    where `-0.25` means the given band is completely muted, and `0.25` means it is doubled. Modifying the gain could
    also change the volume of the output.
    
    Attributes
    ----------
    _bands : `None` or `dict` of (`int`, `float`) items
        Bands.
    
    Class Attributes
    ----------------
    identifier : `int`
        The filter type's internal identifier.
    json_key : `str`
        The key of the filter used when serializing it.
    """
    identifier = FILTER_IDENTIFIER_EQUALIZER
    json_key = LAVALINK_KEY_FILTER_EQUALIZER
    
    __slots__ = ('_bands',)
    
    def __new__(cls, *band_gain_pairs):
        """
        Creates a new equalizer filter.
        
        Parameters
        ----------
        *band_gain_pairs : `tuple` (`int`, `float`)
            band-gain pairs.
        
        TypeError
            If `band_gain_pairs` contains a non-tuple element.
        ValueError
            If a band is out of range [0:14].
        """
        bands = None
        
        for band_gain_pair in band_gain_pairs:
            if not isinstance(band_gain_pair, tuple):
                raise TypeError(
                    f'`band_gain_pairs` can contain `tuple` elements, got '
                    f'{band_gain_pair.__class__.__name__}; {band_gain_pair!r}; band_gain_pairs={band_gain_pairs!r}.'
                )
            
            band, gain = band_gain_pair
            
            if (band < 0) or (band >= LAVALINK_BAND_COUNT):
                raise ValueError(
                    f'`band`, can be in range [0:14], got {band!r}.'
                )
            
            if gain >= 1.0:
                gain = 1.0
            elif gain < -0.25:
                gain = -0.25
            elif gain == 0.0:
                continue
            
            if bands is None:
                bands = {}
            
            bands[band] = gain
        
        
        self = object.__new__(cls)
        self._bands = bands
        return self
    
    
    @copy_docs(Filter.__eq__)
    def __eq__(self, other):
        if type(self) is not type(other):
            return NotImplemented
        
        if self._bands != other._bands:
            return False
        
        return True
    
    
    @copy_docs(Filter.__hash__)
    def __hash__(self):
        hash_value = 0
        
        bands = self._bands
        if (bands is not None):
            for item in bands.items():
                hash_value ^= hash(item)
        
        return hash_value
    
    
    @copy_docs(Filter.__repr__)
    def __repr__(self):
        repr_parts = ['<', self.__class__.__name__]
        
        bands = self._bands
        if (bands is not None):
            repr_parts.append(' bands={')
            field_added = False
            
            for band, gain in sorted(bands.items()):
                if field_added:
                    repr_parts.append(', ')
                else:
                    field_added = True
                
                repr_parts.append(repr(band))
                repr_parts.append(': ')
                repr_parts.append(repr(gain))
            
            repr_parts.append('}')
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    @copy_docs(Filter.__bool__)
    def __bool__(self):
        if (self._bands is not None):
            return True
        
        return False
    
    
    @copy_docs(Filter.to_data)
    def to_data(self):
        bands = self._bands
        if (bands is None):
            return []
        
        return [
            {
                LAVALINK_KEY_FILTER_EQUALIZER_BAND: band,
                LAVALINK_KEY_FILTER_EQUALIZER_GAIN: gain,
            }
            for band, gain in self._bands
        ]


class Karaoke(Filter):
    """
    Uses equalization to eliminate part of a band, usually targeting vocals.
    
    Attributes
    ----------
    _filter_band : `float`
        Filter band.
    _filter_width : `float`
        Filter width.
    _level : `float`
        Effect level.
    _mono_level : `float`
        Effect mono level.
    
    Class Attributes
    ----------------
    identifier : `int`
        The filter type's internal identifier.
    json_key : `str`
        The key of the filter used when serializing it.
    """
    identifier = FILTER_IDENTIFIER_KARAOKE
    json_key = LAVALINK_KEY_FILTER_KARAOKE
    
    __slots__ = ('_filter_band', '_filter_width', '_level', '_mono_level')
    
    def __new__(cls, *, filter_band=220.0, filter_width=100.0, level=1.0, mono_level=1.0):
        """
        Creates a new karaoke filter.
        
        Parameters
        ----------
        filter_band : `float` = `220.0`, Optional (Keyword only)
            Filter band.
            
        filter_width : `float` = `100.0`, Optional (Keyword only)
            Filter width.
        
        level : `float` = `1.0`, Optional (Keyword only)
            Effect level.
        
        mono_level : `float` = `1.0`, Optional (Keyword only)
            Effect mono level.
        
        Raises
        ------
        TypeError
            - If `level` is not `float`.
            - If `mono_level` is not `float`.
            - If `filter_band` is not `float`.
            - If `filter_width` is not `float`.
        """
        if not isinstance(level, float):
            raise TypeError(
                f'`level` can be `float`, got {level.__class__.__name__}; {level!r}.'
            )
        
        if not isinstance(mono_level, float):
            raise TypeError(
                f'`mono_level` can be `float`, got {mono_level.__class__.__name__}; {mono_level!r}.'
            )
        
        if not isinstance(filter_band, float):
            raise TypeError(
                f'`filter_band` can be `float`, got {filter_band.__class__.__name__}; {filter_band!r}.'
            )
        
        if not isinstance(filter_width, float):
            raise TypeError(
                f'`filter_width` can be `float`, got {filter_width.__class__.__name__}; {filter_width!r}.'
            )
        
        self = object.__new__(cls)
        self._filter_band = filter_band
        self._filter_width = filter_width
        self._level = level
        self._mono_level = mono_level
        return self
    
    
    @copy_docs(Filter.__eq__)
    def __eq__(self, other):
        if type(self) is not type(other):
            return NotImplemented
        
        if self._filter_band != other._filter_band:
            return False
        
        if self._filter_width != other._filter_width:
            return False
        
        if self._level != other._level:
            return False
        
        if self._mono_level != other._mono_level:
            return False
        
        return True
    
    
    @copy_docs(Filter.__hash__)
    def __hash__(self):
        hash_value = 0
        hash_value ^= hash(self._filter_band)
        hash_value ^= hash(self._filter_width)
        hash_value ^= hash(self._level)
        hash_value ^= hash(self._mono_level)
        return hash_value
    
    
    @copy_docs(Filter.__repr__)
    def __repr__(self):
        repr_parts = ['<', self.__class__.__name__]
        
        repr_parts.append(' level=')
        repr_parts.append(format(self._level, '.02f'))
        
        repr_parts.append(', mono_level=')
        repr_parts.append(format(self._mono_level, '.02f'))
        
        repr_parts.append(', filter_band=')
        repr_parts.append(format(self._filter_band, '.02f'))
        
        repr_parts.append(', filter_width=')
        repr_parts.append(format(self._filter_width, '.02f'))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
        
    @copy_docs(Filter.__bool__)
    def __bool__(self):
        return True
    
    
    @copy_docs(Filter.to_data)
    def to_data(self):
        return {
            LAVALINK_KEY_FILTER_KARAOKE_FILTER_BAND: self._filter_band,
            LAVALINK_KEY_FILTER_KARAOKE_FILTER_WIDTH: self._filter_width,
            LAVALINK_KEY_FILTER_KARAOKE_LEVEL: self._level,
            LAVALINK_KEY_FILTER_KARAOKE_MONO_LEVEL: self._mono_level,
        }



class LowPass(Filter):
    """
    Higher frequencies get suppressed, while lower frequencies pass through this filter, thus the name low pass.
    
    Attributes
    ----------
    _smoothing : `float`
        Smoothing.
    
    Class Attributes
    ----------------
    identifier : `int`
        The filter type's internal identifier.
    json_key : `str`
        The key of the filter used when serializing it.
    """
    identifier = FILTER_IDENTIFIER_LOW_PASS
    json_key = LAVALINK_KEY_FILTER_LOW_PASS
    
    __slots__ = ('_smoothing',)
    
    def __new__(cls, smoothing):
        """
        Creates a new low pass filter.
        
        Parameters
        ----------
        smoothing : `float`
            Smoothing.
        
        Raises
        ------
        TypeError
            - If `smoothing` is not `float`.
        """
        if not isinstance(smoothing, float):
            raise TypeError(
                f'`smoothing` can be `float`, got {smoothing.__class__.__name__}; {smoothing!r}.'
            )
        
        
        self = object.__new__(cls)
        self._smoothing = smoothing
        return self
    
    
    @copy_docs(Filter.__eq__)
    def __eq__(self, other):
        if type(self) is not type(other):
            return NotImplemented
        
        if self._smoothing != other._smoothing:
            return False
        
        return True
    
    
    @copy_docs(Filter.__hash__)
    def __hash__(self):
        hash_value = 0
        
        smoothing = self._smoothing
        if (smoothing != 20.0):
            hash_value ^= hash(smoothing)
        
        
        return hash_value
    
    
    @copy_docs(Filter.__repr__)
    def __repr__(self):
        repr_parts = ['<', self.__class__.__name__]
        
        repr_parts.append(' smoothing=')
        repr_parts.append(format(self._smoothing, '.02f'))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    @copy_docs(Filter.__bool__)
    def __bool__(self):
        if self._smoothing != 20.0:
            return True
        
        return False
    
    
    @copy_docs(Filter.to_data)
    def to_data(self):
        return {
            LAVALINK_KEY_FILTER_LOW_PASS_SMOOTHING: self._smoothing,
        }



class Rotation(Filter):
    """
    Rotates the sound around the stereo channels/user headphones aka Audio Panning.
    
    It can produce an effect similar to: ``this:https://youtu.be/QB9EB8mTKcc`` (without the reverb).
    
    Attributes
    ----------
    _rotation : `float`
        The frequency of the audio rotating around the listener in seconds.
        
        > 0.2 is similar to the example video above.
    
    Class Attributes
    ----------------
    identifier : `int`
        The filter type's internal identifier.
    json_key : `str`
        The key of the filter used when serializing it.
    """
    identifier = FILTER_IDENTIFIER_ROTATION
    json_key = LAVALINK_KEY_FILTER_ROTATION
    
    __slots__ = ('_rotation',)
    
    def __new__(cls, rotation):
        """
        Creates a new rotation filter.
        
        Parameters
        ----------
        rotation : `float`
            The frequency of the audio rotating around the listener in Hz.
        
        Raises
        ------
        TypeError
            - If `rotation` is not `float`.
            
        """
        if not isinstance(rotation, float):
            raise TypeError(
                f'`rotation` can be `float`, got {rotation.__class__.__name__}; {rotation!r}.'
            )
        
        
        self = object.__new__(cls)
        self._rotation = rotation
        return self
    
    
    @copy_docs(Filter.__eq__)
    def __eq__(self, other):
        if type(self) is not type(other):
            return NotImplemented
        
        if self._rotation != other._rotation:
            return False
        
        return True
    
    
    @copy_docs(Filter.__hash__)
    def __hash__(self):
        hash_value = 0
        
        rotation = self._rotation
        if (rotation != 0.0):
            hash_value ^= hash(rotation)
        
        
        return hash_value
    
    
    @copy_docs(Filter.__repr__)
    def __repr__(self):
        repr_parts = ['<', self.__class__.__name__]
        
        repr_parts.append(' rotation=')
        repr_parts.append(format(self._rotation, '.02f'))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    @copy_docs(Filter.__bool__)
    def __bool__(self):
        if self._rotation != 0.0:
            return True
        
        return False
    
    
    @copy_docs(Filter.to_data)
    def to_data(self):
        return {
            LAVALINK_KEY_FILTER_ROTATION_ROTATION: self._rotation,
        }


class Timescale(Filter):
    """
    Changes the speed, pitch, and rate. All default to 1.
    
    Attributes
    ----------
    _pitch : `float`
        Audio pitch.
    _rate : `float`
        Playback rate.
    _speed : `float`
        Playback speed.
    
    Class Attributes
    ----------------
    identifier : `int`
        The filter type's internal identifier.
    json_key : `str`
        The key of the filter used when serializing it.
    """
    identifier = FILTER_IDENTIFIER_TIMESCALE
    json_key = LAVALINK_KEY_FILTER_TIMESCALE
    
    __slots__ = ('_pitch', '_rate', '_speed')
    
    def __new__(cls, *, pitch=1.0, rate=1.0, speed=1.0):
        """
        Creates a new timescale filter.
        
        Parameters
        ----------
        pitch : `float` = `1.0`, Optional (Keyword only):
            Audio pitch.
        
        rate : `float` = `1.0`, Optional (Keyword only):
            Playback rate.
        
        speed : `float` = `1.0`, Optional (Keyword only)
            Playback speed.
        
        Raises
        ------
        TypeError
            - If `speed` is not `float`.
            - If `pitch` is not `float`.
            - If `rate` is not `float`.
        """
        if not isinstance(speed, float):
            raise TypeError(
                f'`speed` can be `float`, got {speed.__class__.__name__}; {speed!r}.'
            )
        
        if not isinstance(pitch, float):
            raise TypeError(
                f'`pitch` can be `float`, got {pitch.__class__.__name__}; {pitch!r}.'
            )
        
        if not isinstance(rate, float):
            raise TypeError(
                f'`rate` can be `float`, got {rate.__class__.__name__}; {rate!r}.'
            )
        
        if speed < 0:
            raise ValueError(
                f'`speed`, can be in range [0:], got {speed!r}.'
            )
        
        if pitch < 0:
            raise ValueError(
                f'`pitch`, can be in range [0:], got {pitch!r}.'
            )
        
        if rate < 0:
            raise ValueError(
                f'`rate`, can be in range [0:], got {rate!r}.'
            )
        
        self = object.__new__(cls)
        self._pitch = pitch
        self._rate = rate
        self._speed = speed
        return self
    
    
    @copy_docs(Filter.__eq__)
    def __eq__(self, other):
        if type(self) is not type(other):
            return NotImplemented
        
        if self._pitch != other._pitch:
            return False
        
        if self._rate != other._rate:
            return False
        
        if self._speed != other._speed:
            return False
        
        return True
    
    
    @copy_docs(Filter.__hash__)
    def __hash__(self):
        hash_value = 0
        
        pitch = self._pitch
        if (pitch != 1.0):
            hash_value ^= hash(pitch)
        
        rate = self._rate
        if (rate != 1.0):
            hash_value ^= hash(rate)
        
        speed = self._speed
        if (speed != 1.0):
            hash_value ^= hash(speed)
        
        return hash_value
    
    
    @copy_docs(Filter.__repr__)
    def __repr__(self):
        repr_parts = ['<', self.__class__.__name__]
        
        speed = self._speed
        if (speed != 1.0):
            field_added = True
            
            repr_parts.append(' speed=')
            repr_parts.append(format(speed, '.02f'))
        
        else:
            field_added = False
        
        pitch = self._pitch
        if (pitch != 1.0):
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            repr_parts.append(' pitch=')
            repr_parts.append(format(pitch, '.02f'))
        
        rate = self._rate
        if (rate != 1.0):
            if field_added:
                repr_parts.append(',')
            
            repr_parts.append(' rate=')
            repr_parts.append(format(rate, '.02f'))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    @copy_docs(Filter.__bool__)
    def __bool__(self):
        if self._pitch != 1.0:
            return True
        
        if self._rate != 1.0:
            return True
        
        if self._speed != 1.0:
            return True
        
        return False
    
    
    @copy_docs(Filter.to_data)
    def to_data(self):
        return {
            LAVALINK_KEY_FILTER_TIMESCALE_PITCH: self._pitch,
            LAVALINK_KEY_FILTER_TIMESCALE_RATE: self._rate,
            LAVALINK_KEY_FILTER_TIMESCALE_SPEED: self._speed,
        }
    


class Tremolo(Filter):
    """
    Uses amplification to create a shuddering effect, where the volume quickly oscillates.
    
    Example ``here:https://en.wikipedia.org/wiki/File:Fuse_Electronics_Tremolo_MK-III_Quick_Demo.ogv``.
    
    Attributes
    ----------
    _depth : `float`
        Effect depth.
    _frequency : `float`
        Effect frequency.
    
    Class Attributes
    ----------------
    identifier : `int`
        The filter type's internal identifier.
    json_key : `str`
        The key of the filter used when serializing it.
    """
    identifier = FILTER_IDENTIFIER_TREMOLO
    json_key = LAVALINK_KEY_FILTER_TREMOLO
    
    __slots__ = ('_depth', '_frequency')
    
    def __new__(cls, frequency, depth):
        """
        Creates a new tremolo filter.
        
        Parameters
        ----------
        frequency : `float`
            Effect frequency.
        depth : `float`
            Effect depth.
        
        Raises
        ------
        TypeError
            - If `frequency` is not `float`.
            - If `depth` is not `float`.
        ValueError
            - If `frequency` is out of `(0.0:)`.
            - If `depth` is out of `(0.0:1.0]`.
            
        """
        if not isinstance(frequency, float):
            raise TypeError(
                f'`frequency` can be `float`, got {frequency.__class__.__name__}; {frequency!r}.'
            )
        
        if not isinstance(depth, float):
            raise TypeError(
                f'`depth` can be `float`, got {depth.__class__.__name__}; {depth!r}.'
            )
        
        
        if frequency <= 0.0:
            raise ValueError(
                f'`frequency` can be in range `(0.0:)`, got {frequency!r}.'
        )
        
        
        if (depth <= 0.0) or (depth > 1.0):
            raise ValueError(
                f'`depth` can be in range `(0.0:1.0]`, got {depth!r}.'
        )
        
        
        self = object.__new__(cls)
        self._depth = depth
        self._frequency = frequency
        return self
    
    
    @copy_docs(Filter.__eq__)
    def __eq__(self, other):
        if type(self) is not type(other):
            return NotImplemented
        
        if self._depth != other._depth:
            return False
        
        if self._frequency != other._frequency:
            return False
        
        return True
    
    
    @copy_docs(Filter.__hash__)
    def __hash__(self):
        hash_value = 0
        
        depth = self._depth
        if (depth != 0.0):
            hash_value ^= hash(depth)
        
        hash_value ^= hash(self._frequency)
        
        return hash_value
    
    
    @copy_docs(Filter.__repr__)
    def __repr__(self):
        repr_parts = ['<', self.__class__.__name__]
        
        repr_parts.append(' depth=')
        repr_parts.append(format(self._depth, '.02f'))
        
        repr_parts.append(', frequency=')
        repr_parts.append(format(self._frequency, '.02f'))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    @copy_docs(Filter.__bool__)
    def __bool__(self):
        if self._depth != 0.0:
            return True
        
        return False
    
    
    @copy_docs(Filter.to_data)
    def to_data(self):
        return {
            LAVALINK_KEY_FILTER_TREMOLO_DEPTH: self._depth,
            LAVALINK_KEY_FILTER_TREMOLO_FREQUENCY: self._frequency,
        }


class Vibrato(Filter):
    """
    Similar to tremolo. While tremolo oscillates the volume, vibrato oscillates the pitch.
    
    Attributes
    ----------
    _depth : `float`
        Effect depth.
    _frequency : `float`
        Effect frequency.
    
    Class Attributes
    ----------------
    identifier : `int`
        The filter type's internal identifier.
    json_key : `str`
        The key of the filter used when serializing it.
    """
    identifier = FILTER_IDENTIFIER_VIBRATO
    json_key = LAVALINK_KEY_FILTER_VIBRATO
    
    __slots__ = ('_depth', '_frequency')
    
    def __new__(cls, frequency, depth):
        """
        Creates a new vibrato filter.
        
        Parameters
        ----------
        frequency : `float`
            Effect frequency.
        depth : `float`
            Effect depth.
        
        Raises
        ------
        TypeError
            - If `frequency` is not `float`.
            - If `depth` is not `float`.
        ValueError
            - If `frequency` is out of `(0.0:14.0]`.
            - If `depth` is out of `(0.0:1.0]`.
            
        """
        if not isinstance(frequency, float):
            raise TypeError(
                f'`frequency` can be `float`, got {frequency.__class__.__name__}; {frequency!r}.'
            )
        
        if not isinstance(depth, float):
            raise TypeError(
                f'`depth` can be `float`, got {depth.__class__.__name__}; {depth!r}.'
            )
        
        
        if (frequency <= 0.0) or (frequency > 14.0):
            raise ValueError(
                f'`frequency` can be in range `(0.0:14.0]`, got {frequency!r}.'
        )
        
        
        if (depth <= 0.0) or (depth > 1.0):
            raise ValueError(
                f'`depth` can be in range `(0.0:1.0]`, got {depth!r}.'
        )
        
        
        self = object.__new__(cls)
        self._depth = depth
        self._frequency = frequency
        return self
    
    
    @copy_docs(Filter.__eq__)
    def __eq__(self, other):
        if type(self) is not type(other):
            return NotImplemented
        
        if self._depth != other._depth:
            return False
        
        if self._frequency != other._frequency:
            return False
        
        return True
    
    
    @copy_docs(Filter.__hash__)
    def __hash__(self):
        hash_value = 0
        
        depth = self._depth
        if (depth != 0.0):
            hash_value ^= hash(depth)
        
        hash_value ^= hash(self._frequency)
        
        return hash_value
    
    
    @copy_docs(Filter.__repr__)
    def __repr__(self):
        repr_parts = ['<', self.__class__.__name__]
        
        repr_parts.append(' depth=')
        repr_parts.append(format(self._depth, '.02f'))
        
        repr_parts.append(', frequency=')
        repr_parts.append(format(self._frequency, '.02f'))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    @copy_docs(Filter.__bool__)
    def __bool__(self):
        if self._depth != 0.0:
            return True
        
        return False
    
    
    @copy_docs(Filter.to_data)
    def to_data(self):
        return {
            LAVALINK_KEY_FILTER_VIBRATO_DEPTH: self._depth,
            LAVALINK_KEY_FILTER_VIBRATO_FREQUENCY: self._frequency,
        }



class Volume(Filter):
    """
    Float value where `1.0` is 100%.
    
    Values `> 1.0` may cause clipping.
    
    Attributes
    ----------
    _volume : `float`
        The volume to set.
    
    Class Attributes
    ----------------
    identifier : `int`
        The filter type's internal identifier.
    json_key : `str`
        The key of the filter used when serializing it.
    """
    identifier = FILTER_IDENTIFIER_VOLUME
    json_key = LAVALINK_KEY_FILTER_VOLUME
    
    __slots__ = ('_volume',)
    
    def __new__(cls, volume):
        """
        Creates a new volume filter.
        
        Parameters
        ----------
        volume : `float`
        The volume to set.
        
        Raises
        ------
        TypeError
            - If `volume` is not `float`.
        ValueError
            - If volume is over `5`.
        """
        if not isinstance(volume, float):
            raise TypeError(
                f'`volume` can be `float`, got {volume.__class__.__name__}; {volume!r}.'
            )
        
        if volume > 5.0:
            raise ValueError(
                f'`volume` can be in range `[0.0:5.0]`, got {volume!r}.'
        )
        
        self = object.__new__(cls)
        self._volume = volume
        return self
    
    
    @copy_docs(Filter.__eq__)
    def __eq__(self, other):
        if type(self) is not type(other):
            return NotImplemented
        
        if self._volume != other._volume:
            return False
        
        return True
    
    
    @copy_docs(Filter.__hash__)
    def __hash__(self):
        return hash(self._volume)
    
    
    @copy_docs(Filter.__repr__)
    def __repr__(self):
        repr_parts = ['<', self.__class__.__name__]
        
        repr_parts.append(' volume=')
        repr_parts.append(format(self._volume, '.02f'))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    @copy_docs(Filter.__bool__)
    def __bool__(self):
        if self._volume != 1.0:
            return True
        
        return False
    
    
    @copy_docs(Filter.to_data)
    def to_data(self):
        return self._volume



__all__ = ('MessageInteraction',)

from scarletio import include

from ..bases import DiscordEntity
from ..user import User


InteractionType = include('InteractionType')


class MessageInteraction(DiscordEntity):
    """
    Sent with a ``Message``, when the it is a response to an ``InteractionEvent``.
    
    Attributes
    ----------
    id : `int`
        The interaction's identifier.
    name : `str`
        The invoked interaction's name.
    type : ``InteractionType``
        The interaction's type.
    sub_command_name_stack : `None`, `tuple` of `str`
        The sub-command-group and sub-command names.
    user : ``ClientUserBase``
        Who invoked the interaction.
    """
    __slots__ = ('name', 'type', 'sub_command_name_stack', 'user')
    
    def __new__(cls, data, guild_id):
        """
        Creates a new ``MessageInteraction`` from the received data.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Message interaction data.
        guild_id : `int`
            The respective message's guild's identifier.
        """
        full_name = data.get('name', None)
        if full_name is None:
            name = ''
            sub_command_name_stack = None
        
        else:
            name_split = full_name.split(' ')
            
            name_split_length = len(name_split)
            if name_split_length == 1:
                name = name_split[0]
                sub_command_name_stack = None
            
            elif name_split_length > 1:
                name = name_split[0]
                sub_command_name_stack = tuple(name_split[1:])
            
            else:
                name = ''
                sub_command_name_stack = None
        
        self = object.__new__(cls)
        self.id = int(data['id'])
        self.name = name
        self.sub_command_name_stack = sub_command_name_stack
        self.type = InteractionType.get(data['type'])
        self.user = User.from_data(data['user'], data.get('member', None), guild_id)
        
        return self
    
    
    def __repr__(self):
        """Returns the message interaction's representation."""
        repr_parts = ['<', self.__class__.__name__, ' id=', repr(self.id), ', type=']
        
        interaction_type = self.type
        repr_parts.append(interaction_type.name)
        repr_parts.append(' (')
        repr_parts.append(repr(interaction_type.value))
        repr_parts.append(')')
        
        repr_parts.append(', name=')
        repr_parts.append(repr(self.name))
        
        sub_command_name_stack = self.sub_command_name_stack
        if (sub_command_name_stack is not None):
            repr_parts.append(', sub_command_name_stack=')
            repr_parts.append(repr(sub_command_name_stack))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    @property
    def joined_name(self):
        """
        Returns the joined name of the message interaction.
        
        Returns
        -------
        joined_name : `str`
        """
        name = self.name
        sub_command_name_stack = self.sub_command_name_stack
        if (sub_command_name_stack is None):
            return name
        
        return ' '.join([name, *sub_command_name_stack])
    
    
    def to_data(self):
        """
        Tries to convert the message interaction back to json serializable dictionary.
        
        Returns
        -------
        data : `dict` of (`str`, `Any`)
        """
        return {
            'id': str(self.id),
            'name': self.joined_name,
            'type': self.type.value,
            'user': self.user.to_data()
        }



__all__ = ('ComponentInteraction',)

import reprlib, warnings

from scarletio import copy_docs

from ..components import ComponentBase, ComponentType

from .interaction_field_base import InteractionFieldBase


class ComponentInteraction(InteractionFieldBase):
    """
    A component interaction of an ``InteractionEvent``.
    
    Attributes
    ----------
    custom_id : `None`, `str`
        The component's custom identifier.
    options : `None`, `tuple` of `str`
        Option values selected of the respective interaction.
    type : ``ComponentType``
        The component's type.
    """
    __slots__ = ('type', 'custom_id', 'components', 'options')
    
    @copy_docs(InteractionFieldBase.__new__)
    def __new__(cls, data, interaction_event):
        # custom_id
        custom_id = data.get('custom_id', None)
        
        # type
        type_ = ComponentType.get(data['component_type'])
        
        # options
        option_datas = data.get('values', None)
        if (option_datas is None) or (not option_datas):
            options = None
        else:
            options = tuple(option_datas)
        
        self = object.__new__(cls)
        self.custom_id = custom_id
        self.options = options
        self.type = type_
        return self
    
    
    @copy_docs(InteractionFieldBase.__repr__)
    def __repr__(self):
        repr_parts = ['<', self.__class__.__name__]
        
        # type
        type_ = self.type
        repr_parts.append(' type=')
        repr_parts.append(type_.name)
        repr_parts.append(' (')
        repr_parts.append(repr(type_.value))
        repr_parts.append(')')
        
        custom_id = self.custom_id
        if (custom_id is not None):
            repr_parts.append(', custom_id=')
            repr_parts.append(reprlib.repr(custom_id))
        
        options = self.options
        if (options is not None):
            repr_parts.append(', options=[')
            index = 0
            limit = len(options)
            while True:
                option = options[index]
                repr_parts.append(repr(option))
                
                index += 1
                if index == limit:
                    break
                
                repr_parts.append(', ')
                continue
            
            repr_parts.append(']')
        
        repr_parts.append('>')
        
        return ''.join(repr_parts)
    
    
    @copy_docs(InteractionFieldBase.__eq__)
    def __eq__(self, other):
        other_type = type(other)
        # Compare with self type.
        if other_type is type(self):
            # type
            if self.type is not other.type:
                return False
            #custom_id
            if self.custom_id != other.custom_id:
                return False
            
            return True
        
        # Compare with components.
        if issubclass(other_type, ComponentBase):
            # Check `type` before `custom_id`
            
            # type
            if self.type is not other.type:
                return False
            
            # custom_id
            if self.custom_id != other.custom_id:
                return False
            
            return True
        
        
        return NotImplemented
    
    
    @copy_docs(InteractionFieldBase.__hash__)
    def __hash__(self):
        hash_value = 0
        
        # type
        hash_value ^= self.type.value
        
        # custom_id
        hash_value ^= hash(self.custom_id)
        
        # options
        options = self.options
        if (options is not None):
            hash_value ^ len(options) << 8
            for option in options:
                hash_value ^ hash(option)
        
        return hash_value



__all__ = ()

from .shared import convert_nothing, convert_snowflake_array


EMOJI_CONVERTERS = {
    'available': convert_nothing,
    'name': convert_nothing,
    'role_ids': convert_snowflake_array,
}



__all__ = ('GetTracksResult', 'Track', )

from base64 import b64decode
from math import floor

from scarletio import RichAttributeErrorBaseType, un_map_pack

from .constants import (
    LAVALINK_KEY_END_TIME, LAVALINK_KEY_PLAYLIST, LAVALINK_KEY_PLAYLIST_NAME,
    LAVALINK_KEY_PLAYLIST_SELECTED_TRACK_INDEX, LAVALINK_KEY_START_TIME, LAVALINK_KEY_TRACK, LAVALINK_KEY_TRACKS,
    LAVALINK_KEY_TRACK_AUTHOR, LAVALINK_KEY_TRACK_BASE64, LAVALINK_KEY_TRACK_DICT, LAVALINK_KEY_TRACK_DURATION_MS,
    LAVALINK_KEY_TRACK_IDENTIFIER, LAVALINK_KEY_TRACK_IS_STREAM, LAVALINK_KEY_TRACK_POSITION_MS,
    LAVALINK_KEY_TRACK_SEEKABLE, LAVALINK_KEY_TRACK_TITLE, LAVALINK_KEY_TRACK_URL
)


def _iter_configured_track_attributes(configured_track):
    """
    Yields `key` - `value` pairs of a configured track.
    
    This function is an iterable generator.
    
    Parameters
    ----------
    configured_track : ``ConfiguredTrack``
        The configured pack to deserialize.
    
    Yields
    ------
    key : `str`
    value : `Any`
    """
    yield LAVALINK_KEY_TRACK, configured_track.track.base64
    
    start_time = configured_track.start_time
    if start_time:
        yield LAVALINK_KEY_START_TIME, floor(start_time * 1000.0)
    
    end_time = configured_track.end_time
    if end_time:
        yield LAVALINK_KEY_END_TIME, floor(end_time * 1000.0)


class Track(RichAttributeErrorBaseType):
    """
    Represents an audio track sent by lavalink.
    
    Parameters
    ----------
    author : `str`
        The track's uploader.
    base64 : `str`
        The base 64 version of the track.
    duration : `float`
        The duration of the track in seconds.
    identifier : `str`
        The track's identifier.
    is_stream : `bool`
        Whether the track is a live-stream.
    is_seekable : `bool`
        Whether the track supports seeking.
    position : `float`
        The position of the track in seconds.
    title : `str`
        The track's title.
    url : `None`, `str`
        The url of the track.
    """
    __slots__ = ('author', 'base64', 'duration', 'identifier', 'is_seekable', 'is_stream', 'position', 'title', 'url')
    
    def __new__(cls, data):
        """
        Creates a new track from the given data.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Track data.
        """
        self = object.__new__(cls)
        self.base64 = data[LAVALINK_KEY_TRACK_BASE64]
        
        track_data = data[LAVALINK_KEY_TRACK_DICT]
        
        self.author = track_data[LAVALINK_KEY_TRACK_AUTHOR]
        self.duration = track_data[LAVALINK_KEY_TRACK_DURATION_MS]*0.001
        self.identifier = track_data[LAVALINK_KEY_TRACK_IDENTIFIER]
        self.is_stream = track_data[LAVALINK_KEY_TRACK_IS_STREAM]
        self.is_seekable = track_data[LAVALINK_KEY_TRACK_SEEKABLE]
        self.title = track_data[LAVALINK_KEY_TRACK_TITLE]
        self.url = track_data[LAVALINK_KEY_TRACK_URL]
        
        try:
            position = track_data[LAVALINK_KEY_TRACK_POSITION_MS]
        except KeyError:
            position = 0.0
        else:
            position *= 0.001
        self.position = position
        
        return self
    
    @classmethod
    def from_base64(cls, base64):
        """
        Creates a new track from base 64 data.
        
        Parameters
        ----------
        base64 : `str`
            Track data in base64.
        """
        data = b64decode(base64)
        
        # read flags
        
        flags = int.from_bytes(data[0:4], 'big')
        flags = (flags & 0xC0000000) >> 30
        
        # read version
        
        if flags & 1:
            version = data[4]
            cursor_start = 5
        else:
            version = 1
            cursor_start = 4
        
        # read title
        
        cursor_end = cursor_start + 2
        title_length = int.from_bytes(data[cursor_start:cursor_end], 'big')
        
        cursor_start = cursor_end
        cursor_end = cursor_start + title_length
        title = data[cursor_start:cursor_end]
        title = title.decode(errors='ignore')
        
        # read author
        
        cursor_start = cursor_end
        cursor_end = cursor_start + 2
        author_length = int.from_bytes(data[cursor_start:cursor_end], 'big')
        
        cursor_start = cursor_end
        cursor_end = cursor_start + author_length
        author = data[cursor_start:cursor_end]
        author = author.decode(errors='ignore')
        
        # read duration
        
        cursor_start = cursor_end
        cursor_end = cursor_start + 8
        duration = int.from_bytes(data[cursor_start:cursor_end], 'big')
        duration *= 0.001
        
        # read identifier
        
        cursor_start = cursor_end
        cursor_end = cursor_start + 2
        identifier_length = int.from_bytes(data[cursor_start:cursor_end], 'big')
        
        cursor_start = cursor_end
        cursor_end = cursor_start + identifier_length
        identifier = data[cursor_start:cursor_end]
        identifier = identifier.decode()
        
        # read is_stream
        
        if data[cursor_end]:
            is_stream = True
        else:
            is_stream = False
        cursor_end += 1
        
        # read url
        
        if data[cursor_end]:
            cursor_start = cursor_end + 1
            cursor_end = cursor_start + 2
            
            url_length = int.from_bytes(data[cursor_start:cursor_end], 'big')
            
            cursor_start = cursor_end
            cursor_end = cursor_start + url_length
            url = data[cursor_start:cursor_end]
            url = url.decode()
            
        else:
            cursor_end += 1
            url = None
        
        # read source
        
        cursor_start = cursor_end
        cursor_end = cursor_start + 2
        source_length = int.from_bytes(data[cursor_start:cursor_end], 'big')
        
        cursor_start = cursor_end
        cursor_end = cursor_start + source_length
        source = data[cursor_start:cursor_end]
        source = source.decode()
        
        # read position
        
        cursor_start = cursor_end
        cursor_end = cursor_start + 8
        position = int.from_bytes(data[cursor_start:cursor_end], 'big')
        position *= 0.001
        
        self = object.__new__(cls)
        
        self.base64 = base64
        self.author = author
        self.duration = duration
        self.identifier = identifier
        self.is_stream = is_stream
        self.is_seekable = (not is_stream)
        self.title = title
        self.url = url
        self.position = position
        
        return self
    
    
    def __repr__(self):
        """Returns the track's representation."""
        repr_parts = ['<', self.__class__.__name__]
        
        if self.is_stream:
            field_added = True
            repr_parts.append(' stream')
        else:
            field_added = False
        
        if self.is_seekable:
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' seekable')
        
        if not field_added:
            repr_parts.append(',')
        
        repr_parts.append(' author=')
        repr_parts.append(repr(self.author))
        
        repr_parts.append(', duration=')
        repr_parts.append(self.duration.__format__('.03f'))
        repr_parts.append('s')
        
        position = self.position
        if position:
            repr_parts.append(', position=')
            repr_parts.append(position.__format__('.03f'))
            repr_parts.append('s')
        
        repr_parts.append(', identifier=')
        repr_parts.append(repr(self.identifier))
        
        repr_parts.append(', title=')
        repr_parts.append(repr(self.title))
        
        url = self.url
        if (url is not None):
            repr_parts.append(', url=')
            repr_parts.append(repr(url))
        
        repr_parts.append('>')
        
        return ''.join(repr_parts)
    
    
    def __hash__(self):
        """Returns the track's hash."""
        return hash(self.base64)
    
    
    def __eq__(self, other):
        """Returns whether the two tracks are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        if self.base64 == other.base64:
            return True
        
        return False


class ConfiguredTrack(RichAttributeErrorBaseType):
    """
    Represents a track added to a player. Not like generic tracks, this supports additional attributes as well.
    
    Attributes
    ----------
    _added_attributes : `None`, `dict` of (`str`, `Any`) items
        Additionally passed attributes when registering a track.
    
    end_time : `float`
        When the track will end.
        
        Defaults to `0.0`.
    
    start_time : `float`
        When the track will start.
        
        Defaults to `0.0`.
    
    track : ``Track``
        The source track.
    """
    __slots__ = ('_added_attributes', 'end_time', 'start_time', 'track')
    
    def __new__(cls, track, start_time, end_time, added_attributes):
        """
        Creates a new configured track from the given parameters.
        
        Parameters
        ----------
        track : ``Track``
            The wrapped track.
        start_time : `float`
            Where the track will start in seconds.
        end_time : `float`
            Where the track will start in seconds.
        added_attributes : `dict` of (`str`, `Any`)
            Additional user defined attributes.
        
        Raises
        ------
        TypeError
            If `track` is neither ``Track``, nor ``ConfiguredTrack``.
        ValueError
            - If `start_time` is out of the expected [0.0:duration] range.
            - If `end_time` is out of the expected [0.0:duration] range.
        """
        if isinstance(track, Track):
            configuration = None
        elif isinstance(track, ConfiguredTrack):
            configuration = track
            track = track.track
        else:
            raise TypeError(
                f'`track` can be `{Track.__name__}`, `{ConfiguredTrack.__name__}`, got '
                f'{track.__class__.__name__}; {track!r}.'
            )
        
        duration = track.duration
        if start_time:
            if (start_time < 0.0) or (start_time > duration):
                raise ValueError(
                    f'`start_time` can be in range [0.0:{duration:.3f}], got {end_time:.3f}.'
                )
        else:
            if (configuration is not None):
                start_time = configuration.start_time
        
        if end_time:
            if (end_time < 0.0) or (end_time > duration):
                raise ValueError(
                    f'`end_time` can be in range [0.0:{duration:.3f}], got {end_time:.3f}.'
                )
        else:
            if (configuration is not None):
                start_time = configuration.end_time
        
        if added_attributes:
            if (configuration is not None):
                configuration_added_attributes = configuration._added_attributes
                if (configuration_added_attributes is not None):
                    added_attributes = {**configuration_added_attributes, **added_attributes}
        else:
            if (configuration is None):
                added_attributes = None
            else:
                added_attributes = configuration._added_attributes
                if (added_attributes is not None):
                    added_attributes = added_attributes.copy()
        
        self = object.__new__(cls)
        self.track = track
        self.start_time = start_time
        self.end_time = end_time
        self._added_attributes = added_attributes
    
        return self
    
    
    def __getattr__(self, attribute_name):
        """Tries to find the attributes from the added attributes."""
        added_attributes = self._added_attributes
        if (added_attributes is not None):
            try:
                return added_attributes[attribute_name]
            except KeyError:
                pass
        
        RichAttributeErrorBaseType.__getattr__(self, attribute_name)
    
    
    def __dir__(self):
        """Returns the attribute names of the object."""
        directory = set(object.__dir__(self))
        added_attributes = self._added_attributes
        if (added_attributes is not None):
            directory.update(added_attributes.keys())
        
        return sorted(directory)
    
    
    def un_pack(self):
        """
        Unpacks the configured track to `key` - `value` pairs accepted by lavalink.
        
        Returns
        -------
        un_map_packer : ``un_map_pack``
        """
        return un_map_pack(_iter_configured_track_attributes(self))
    
    
    def __repr__(self):
        """Returns the configured track's representation."""
        repr_parts = [
            '<', self.__class__.__name__,
            ' track=', repr(self.track),
        ]
        
        start_time = self.start_time
        if start_time:
            repr_parts.append(', start_time=')
            repr_parts.append(start_time.__format__('.3f'))
            repr_parts.append('s')
        
        end_time = self.end_time
        if end_time:
            repr_parts.append(', end_time=')
            repr_parts.append(end_time.__format__('.3f'))
            repr_parts.append('s')
        
        added_attributes = self._added_attributes
        if (added_attributes is not None):
            for attribute_name, attribute_value in added_attributes.items():
                repr_parts.append(', ')
                repr_parts.append(attribute_name)
                repr_parts.append('=')
                repr_parts.append(repr(attribute_value))
        
        repr_parts.append('>')
        
        return ''.join(repr_parts)
    
    
    def __hash__(self):
        """Returns the configured track's hash value."""
        hash_value = 0
        
        start_time = self.start_time
        if start_time:
            hash_value ^= (1 << 7)
            hash_value ^= hash(start_time)
        
        end_time = self.end_time
        if end_time:
            hash_value ^= (1 << 14)
            hash_value ^= hash(end_time)
        
        hash_value ^= hash(self.track)
        
        added_attributes = self.added_attributes
        if (added_attributes is not None):
            hash_value ^= len(added_attributes)
            
            for item in added_attributes.items():
                hash_value ^= hash(item)
        
        return hash_value
    
    
    def __eq__(self, other):
        """Returns whether the two configured tracks are the same."""
        if type(self) is not type(other):
            return NotImplemented
        
        if self.track != other.track:
            return False
        
        if self.end_time != other.end_time:
            return False
        
        if self.start_time != other.start_time:
            return False
        
        if self._added_attributes != other._added_attributes:
            return False
        
        return True
    
    
    @property
    def author(self):
        """
        Returns the track's uploader.
        
        Returns
        -------
        author : `str`
        """
        return self.track.author
    
    
    @property
    def base64(self):
        """
        Returns the track's base 64 version.
        
        Returns
        -------
        base64 : `str`
        """
        return self.track.base64
    
    
    @property
    def duration(self):
        """
        Returns the track's duration in seconds.
        
        Returns
        -------
        duration : `float`
        """
        return self.track.duration
    
    
    @property
    def identifier(self):
        """
        Returns the track's identifier.
        
        Returns
        -------
        identifier : `str`
        """
        return self.track.identifier
    
    
    @property
    def is_stream(self):
        """
        Returns whether the track is a live-stream.
        
        Returns
        -------
        is_stream : `bool`
        """
        return self.track.is_stream
    
    
    @property
    def is_seekable(self):
        """
        Returns whether the track supports seeking.
        
        Returns
        -------
        is_seekable : `bool`
        """
        return self.track.is_seekable
    
    
    @property
    def position(self):
        """
        Returns the position of the track.
        
        Returns
        -------
        position : `float`
        """
        return self.track.position
    
    
    @property
    def title(self):
        """
        Returns the track's title.
        
        Returns
        -------
        title : `str`
        """
        return self.track.title
    
    
    @property
    def url(self):
        """
        Returns the track's url.
        
        Returns
        -------
        url : `None`, `str`
        """
        return self.track.url
    
    
    def copy(self):
        """
        Copies the configured track.
        
        Returns
        -------
        new : ``ConfiguredTrack``
        """
        added_attributes = self._added_attributes
        if (added_attributes is not None):
            added_attributes = added_attributes.copy()
        
        new = object.__new__(type(self))
        
        new._added_attributes = added_attributes
        new.end_time = self.end_time
        new.start_time = self.start_time
        new.track = self.track
        
        return new


class GetTracksResult(RichAttributeErrorBaseType):
    """
    Returned by ``SolarClient.get_tracks`` if the request succeeded.
    
    Attributes
    ----------
    playlist_name : `None`, `str`
        The playlist's name, of what the tracks are part.
        
        Defaults to `None`.
        
    selected_track_index : `int`
        The selected track's index inside of the playlist.
        
        Defaults to `-1`.
    
    tracks : `None`, `tuple` of ``Track``
        The matched tracks.
        
        Defaults to `None`.
    """
    __slots__ = ('playlist_name', 'selected_track_index', 'tracks',)
    
    def __new__(cls, data):
        """
        Creates a new get tracks result instance from the given response data.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Response data.
        """
        playlist_data = data[LAVALINK_KEY_PLAYLIST]
        if playlist_data is None:
            playlist_name = None
            selected_track_index = -1
        else:
            playlist_name = playlist_data.get(LAVALINK_KEY_PLAYLIST_NAME, None)
            selected_track_index = playlist_data.get(LAVALINK_KEY_PLAYLIST_SELECTED_TRACK_INDEX, -1)
        
        track_datas = data[LAVALINK_KEY_TRACKS]
        if (track_datas is None) or (not track_datas):
            tracks = None
        else:
            tracks = tuple(Track(track_data) for track_data in track_datas)
        
        self = object.__new__(cls)
        self.playlist_name = playlist_name
        self.selected_track_index = selected_track_index
        self.tracks = tracks
        return self
    
    
    def __repr__(self):
        """Returns the get tracks result's representation."""
        repr_parts = ['<', self.__class__.__name__]
        
        playlist_name = self.playlist_name
        if (playlist_name is None):
            field_added = False
        else:
            field_added = True
            
            repr_parts.append(' playlist_name=')
            repr_parts.append(repr(playlist_name))
        
        
        selected_track_index = self.selected_track_index
        if (selected_track_index != -1):
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' selected_track_index=')
            repr_parts.append(repr(selected_track_index))
        
        
        tracks = self.tracks
        if (tracks is not None):
            if field_added:
                repr_parts.append(',')
            
            repr_parts.append(' tracks=[')
            
            index = 0
            length = len(tracks)
            
            while True:
                track = tracks[index]
                repr_parts.append(repr(track))
                
                index += 1
                if index == length:
                    break
                
                repr_parts.append(', ')
                continue
            
            repr_parts.append(']')
        
        repr_parts.append('>')
        
        return ''.join(repr_parts)
    
    
    def __len__(self):
        """Returns the length of the tracks."""
        tracks = self.tracks
        if tracks is None:
            length = 0
        else:
            length = len(tracks)
        
        return length
    
    
    def __bool__(self):
        """Returns whether the result has any tracks."""
        tracks = self.tracks
        if tracks is None:
            has_tracks = True
        else:
            if tracks:
                has_tracks = True
            else:
                has_tracks = False
        
        return has_tracks
    
    
    def __getitem__(self, index):
        """Returns the track of the given index"""
        tracks = self.tracks
        if tracks is None:
            raise IndexError(index)
        
        return tracks[index]
    
    
    def __hash__(self):
        """Returns the get tracks result's hash."""
        hash_value = 0
        
        playlist_name = self.playlist_name
        if (playlist_name is not None):
            hash_value ^= hash(playlist_name)
        
        selected_track_index = self.selected_track_index
        if (selected_track_index != -1):
            hash_value ^= (selected_track_index + 1)
        
        tracks = self.tracks
        if (tracks is not None):
            hash_value ^= (len(tracks) << 8)
            
            for track in tracks:
                hash_value ^= hash(track)
        
        return hash_value
    
    
    def __eq__(self, other):
        """Returns whether the two results are the same."""
        if type(self) is not type(other):
            return NotImplemented
        
        if self.playlist_name != other.playlist_name:
            return False
        
        if self.selected_track_index != other.selected_track_index:
            return False
        
        if self.tracks != other.tracks:
            return False
        
        return True



__all__ = ('get_channel_type_name', 'get_channel_type_names',)

__doc__ = """
Contains channel type identifiers.

+---------------------------------------+-------+
| Name                                  | Value |
+=======================================+=======+
| guild_text                            | 0     |
+---------------------------------------+-------+
| private                               | 1     |
+---------------------------------------+-------+
| guild_voice                           | 2     |
+---------------------------------------+-------+
| private_group                         | 3     |
+---------------------------------------+-------+
| guild_category                        | 4     |
+---------------------------------------+-------+
| guild_announcements                   | 5     |
+---------------------------------------+-------+
| guild_store                           | 6     |
+---------------------------------------+-------+
| thread                                | 9     |
+---------------------------------------+-------+
| guild_thread_announcements            | 10    |
+---------------------------------------+-------+
| guild_thread_public                   | 11    |
+---------------------------------------+-------+
| guild_thread_private                  | 12    |
+---------------------------------------+-------+
| guild_stage                           | 13    |
+---------------------------------------+-------+
| guild_directory                       | 14    |
+---------------------------------------+-------+
| guild_forum                           | 15    |
+---------------------------------------+-------+

In addition also extra groups are defined:

+---------------------------------------+-------------------------------+
| Name                                  | Elements                      |
+=======================================+===============================+
| GROUP_MESSAGEABLE                     | guild_text,                   |
|                                       | private,                      |
|                                       | guild_voice,                  |
|                                       | private_group,                |
|                                       | guild_announcements,          |
|                                       | guild_thread_announcements,   |
|                                       | guild_thread_public,          |
|                                       | guild_thread_private          |
+---------------------------------------+-------------------------------+
| GROUP_GUILD_MESSAGEABLE               | guild_text,                   |
|                                       | guild_voice,                  |
|                                       | guild_announcements,          |
|                                       | guild_thread_announcements,   |
|                                       | guild_thread_public,          |
|                                       | guild_thread_private          |
+---------------------------------------+-------------------------------+
| GROUP_GUILD_MAIN_TEXT                 | guild_text,                   |
|                                       | guild_announcements           |
+---------------------------------------+-------------------------------+
| GROUP_CONNECTABLE                     | private,                      |
|                                       | guild_voice,                  |
|                                       | private_group,                |
|                                       | guild_stage                   |
+---------------------------------------+-------------------------------+
| GROUP_GUILD_CONNECTABLE               | guild_voice,                  |
|                                       | guild_stage                   |
+---------------------------------------+-------------------------------+
| GROUP_PRIVATE                         | private,                      |
|                                       | private_group                 |
+---------------------------------------+-------------------------------+
| GROUP_GUILD                           | guild_text,                   |
|                                       | guild_voice,                  |
|                                       | guild_category,               |
|                                       | guild_announcements,          |
|                                       | guild_store,                  |
|                                       | guild_thread_announcements,   |
|                                       | guild_thread_public,          |
|                                       | guild_thread_private,         |
|                                       | guild_stage,                  |
|                                       | guild_directory,              |
|                                       | guild_forum                   |
+---------------------------------------+-------------------------------+
| GROUP_THREAD                          | guild_thread_announcements,   |
|                                       | guild_thread_public,          |
|                                       | guild_thread_private          |
+---------------------------------------+-------------------------------+
| GROUP_CAN_CONTAIN_THREADS             | guild_text,                   |
|                                       | guild_announcements,          |
|                                       | guild_forum                   |
+---------------------------------------+-------------------------------+
| GROUP_IN_PRODUCTION                   | guild_text,                   |
|                                       | private,                      |
|                                       | guild_voice,                  |
|                                       | private_group,                |
|                                       | guild_category,               |
|                                       | guild_announcements,          |
|                                       | guild_thread_announcements,   |
|                                       | guild_thread_public,          |
|                                       | guild_thread_private,         |
|                                       | guild_stage,                  |
|                                       | guild_directory,              |
|                                       | guild_forum                   |
+---------------------------------------+-------------------------------+
| GROUP_CAN_CREATE_INVITE_TO            | guild_text,                   |
|                                       | guild_voice,                  |
|                                       | private_group,                |
|                                       | guild_announcements,          |
|                                       | guild_store,                  |
|                                       | guild_stage,                  |
|                                       | guild_directory               |
+---------------------------------------+-------------------------------+
| GROUP_GUILD_MOVABLE                   | guild_text,                   |
|                                       | guild_voice,                  |
|                                       | guild_category,               |
|                                       | guild_announcements,          |
|                                       | guild_store,                  |
|                                       | guild_stage,                  |
|                                       | guild_directory,              |
|                                       | guild_forum                   |
+---------------------------------------+-------------------------------+
"""
guild_text = 0
private = 1
guild_voice = 2
private_group = 3
guild_category = 4
guild_announcements = 5
guild_store = 6
# 7? Not in use
# 8? Not in use
thread = 9 # Not in use
guild_thread_announcements = 10
guild_thread_public = 11
guild_thread_private = 12
guild_stage = 13
guild_directory = 14
guild_forum = 15


GROUP_MESSAGEABLE = frozenset((
    guild_text,
    private,
    guild_voice,
    private_group,
    guild_announcements,
    guild_thread_announcements,
    guild_thread_public,
    guild_thread_private,
))


GROUP_GUILD_MESSAGEABLE = frozenset((
    guild_text,
    guild_voice,
    guild_announcements,
    guild_thread_announcements,
    guild_thread_public,
    guild_thread_private,
))


GROUP_GUILD_MAIN_TEXT = frozenset((
    guild_text,
    guild_announcements,
))


GROUP_CONNECTABLE = frozenset((
    private,
    guild_voice,
    private_group,
    guild_stage,
))


GROUP_GUILD_CONNECTABLE = frozenset((
    guild_voice,
    guild_stage,
))


GROUP_PRIVATE = frozenset((
    private,
    private_group,
))


GROUP_GUILD = frozenset((
    guild_text,
    guild_voice,
    guild_category,
    guild_announcements,
    guild_store,
    guild_thread_announcements,
    guild_thread_public,
    guild_thread_private,
    guild_stage,
    guild_directory,
    guild_forum,
))


GROUP_THREAD = frozenset((
    guild_thread_announcements,
    guild_thread_public,
    guild_thread_private,
))


GROUP_CAN_CONTAIN_THREADS = frozenset((
    guild_text,
    guild_announcements,
    guild_forum,
))


GROUP_IN_PRODUCTION = frozenset((
    guild_text,
    private,
    guild_voice,
    private_group,
    guild_category,
    guild_announcements,
    guild_thread_announcements,
    guild_thread_public,
    guild_thread_private,
    guild_stage,
    guild_directory,
    guild_forum,
))


GROUP_CAN_CREATE_INVITE_TO = frozenset((
    guild_text,
    guild_voice,
    private_group,
    guild_announcements,
    guild_store,
    guild_stage,
    guild_directory,
    
))

GROUP_GUILD_MOVABLE = frozenset((
    guild_text,
    guild_voice,
    guild_category,
    guild_announcements,
    guild_store,
    guild_stage,
    guild_directory,
    guild_forum,
))


CHANNEL_TYPE_NAMES = {
    guild_text: 'guild text',
    private: 'guild private',
    guild_voice: 'guild voice',
    private_group: 'private group',
    guild_category: 'guild category',
    guild_announcements: 'guild announcements',
    guild_store: 'guild store',
    thread: 'thread',
    guild_thread_announcements: 'guild thread announcements',
    guild_thread_public: 'guild thread public',
    guild_thread_private: 'guild thread private',
    guild_stage: 'guild stage',
    guild_directory: 'guild directory',
    guild_forum: 'guild forum',
}


DEFAULT_CHANNEL_TYPE_NAME = 'unknown'


def get_channel_type_name(channel_type):
    """
    Returns the channel type's name.
    
    Parameters
    ----------
    channel_type : `int`
        The channel type to get it's name.
    
    Returns
    -------
    channel_name : `str`
    """
    return CHANNEL_TYPE_NAMES.get(channel_type, DEFAULT_CHANNEL_TYPE_NAME)


def get_channel_type_names(channel_types):
    """
    Returns multiple channel type's names connected by comma.
    
    Parameters
    ----------
    channel_types : `iterable` of `int`
        The channel types to get the their name.
    
    Returns
    -------
    channel_names : `str`
    """
    return ', '.join(sorted(get_channel_type_name(channel_type) for channel_type in channel_types))



__all__ = ()

from .preinstanced import Locale


DEFAULT_LOCALE = Locale.english_us


def get_locale(value):
    """
    Gets the locale of the given value.
    
    Parameters
    ----------
    value : `None`, `str`
        If given as `None`, returns the default locale.
    
    Returns
    -------
    locale : ``Locale``
    """
    if value is None:
        locale = DEFAULT_LOCALE
    else:
        locale = Locale.get(value)
    
    return locale


def build_locale_dictionary(dictionary):
    """
    Builds a locale dictionary where they keys are all ``Locale``-s.
    
    Parameters
    ----------
    dictionary : `dict` of (`str`, `Any`) items
        The dictionary to process.
    
    Returns
    -------
    transformed : `dict` of (``Locale``, `Any`) items
    """
    if dictionary is not None:
        return {Locale.get(key): value for key, value in dictionary.items()}


def destroy_locale_dictionary(dictionary):
    """
    Builds a json serializable dictionary where they keys are all `str`-s.
    
    Parameters
    ----------
    dictionary : `dict` of (`Locale`, `Any`) items
        The dictionary to process.
    
    Returns
    -------
    transformed : `dict` of (`str`, `Any`) items
    """
    if dictionary is not None:
        return {key.value: value for key, value in dictionary.items()}



__all__ = ()

from ....channel import ChannelFlag, VideoQualityMode
from ....permission import PermissionOverwrite

from ...preinstanced import VoiceRegion

from ..audit_log_change import AuditLogChange

from .shared import _convert_preinstanced, convert_nothing, convert_snowflake


def convert_bool__open(name, data):
    before = data.get('old_value', None)
    if (before is not None):
        before = not before
    
    after = data.get('new_value', None)
    if (after is not None):
        after = not after
    
    return AuditLogChange('open', before, after)


def convert_channel_flags(name, data):
    before = data.get('old_value', None)
    if (before is not None):
        before = ChannelFlag(before)
    
    after = data.get('new_value', None)
    if (after is not None):
        after = ChannelFlag(after)
    
    return AuditLogChange(name, before, after)


def convert_int__auto_archive_after(name, data):
    before = data.get('old_value', None)
    if (before is not None):
        before *= 60
    
    after = data.get('new_value', None)
    if (after is not None):
        after *= 60
    
    return AuditLogChange('auto_archive_after', before, after)


def convert_int__default_auto_archive_after(name, data):
    before = data.get('old_value', None)
    if (before is not None):
        before *= 60
    
    after = data.get('new_value', None)
    if (after is not None):
        after *= 60
    
    return AuditLogChange('default_auto_archive_after', before, after)


def convert_int__slowmode(name, data):
    return convert_nothing('slowmode', data)


def convert_overwrites(name, data):
    before = data.get('old_value', None)
    if (before is not None):
        before = [PermissionOverwrite(overwrite_data) for overwrite_data in before]
    
    after = data.get('new_value', None)
    if (after is not None):
        after = [PermissionOverwrite(overwrite_data) for overwrite_data in after]
    
    return AuditLogChange('overwrites', before, after)


def convert_video_quality_mode(name, data):
    return _convert_preinstanced(name, data, VideoQualityMode)


def convert_voice_region(name, data):
    return _convert_preinstanced('region', data, VoiceRegion)


CHANNEL_CONVERTERS = {
    'archived': convert_nothing,
    'auto_archive_duration': convert_int__auto_archive_after,
    'bitrate': convert_nothing,
    'default_auto_archive_duration': convert_int__default_auto_archive_after,
    'flags': convert_channel_flags,
    'invitable': convert_nothing,
    'locked': convert_bool__open,
    'name': convert_nothing,
    'nsfw': convert_nothing,
    'rate_limit_per_user': convert_int__slowmode,
    'parent_id': convert_snowflake,
    'permission_overwrites': convert_overwrites,
    'position': convert_nothing,
    'rtc_region': convert_voice_region,
    'topic': convert_nothing,
    'type': convert_nothing,
    'video_quality_mode': convert_video_quality_mode,
    'user_limit': convert_nothing,
}



__all__ = ()

import warnings

from scarletio import copy_docs, include

from ...utils.module_deprecation import deprecated_import

from . import channel_types as CHANNEL_TYPES


Channel = include('Channel')

class DeprecatedChannelMetType(type):
    """
    Meta type for the old deprecated channels warning about deprecation when instance or subtype checking.
    """
    def __instancecheck__(cls, instance):
        warnings.warn(
            f'`{cls.__name__}` is deprecated and will be removed in 2022. Please use `{Channel.__name__}` instead.',
            FutureWarning,
            stacklevel = 2,
        )
        
        return isinstance(instance, Channel) and instance.type in cls.allowed_types
    
    def __subclasscheck__(cls, klass):
        warnings.warn(
            f'`{cls.__name__}` is deprecated and will be removed in 2022. Please use `{Channel.__name__}` instead.',
            FutureWarning,
            stacklevel = 2,
        )
        
        return issubclass(klass, Channel) or (klass is cls)


@deprecated_import
class ChannelBase(metaclass=DeprecatedChannelMetType):
    """
    Deprecated and will be removed in 2022 August. Please use the ``Channel`` type instead.
    """
    allowed_types = frozenset((
        CHANNEL_TYPES.guild_text,
        CHANNEL_TYPES.private,
        CHANNEL_TYPES.guild_voice,
        CHANNEL_TYPES.private_group,
        CHANNEL_TYPES.guild_category,
        CHANNEL_TYPES.guild_announcements,
        CHANNEL_TYPES.guild_store,
        CHANNEL_TYPES.thread,
        CHANNEL_TYPES.guild_thread_announcements,
        CHANNEL_TYPES.guild_thread_public,
        CHANNEL_TYPES.guild_thread_private,
        CHANNEL_TYPES.guild_stage,
        CHANNEL_TYPES.guild_directory,
        CHANNEL_TYPES.guild_forum,
    ))
    
    type = 0
    
    @classmethod
    def precreate(cls, channel_id, **kwargs):
        channel_type = next(iter(cls.allowed_types))
        
        warnings.warn(
            (
                f'`{cls.__name__}.precreate(...)` is deprecated and will be removed in 2022 August. Please use: '
                f'`{Channel.__name__}.precreate(channel_id, channel_type={channel_type}, ...)` instead.'
            ),
            FutureWarning,
            stacklevel = 2,
        )
        
        return Channel.precreate(channel_id, channel_type=channel_type, **kwargs)


@deprecated_import
@copy_docs(ChannelBase)
class ChannelTextBase(ChannelBase):
    allowed_types = frozenset((
        CHANNEL_TYPES.guild_text,
        CHANNEL_TYPES.private,
        CHANNEL_TYPES.private_group,
        CHANNEL_TYPES.guild_announcements,
        CHANNEL_TYPES.thread,
        CHANNEL_TYPES.guild_thread_announcements,
        CHANNEL_TYPES.guild_thread_public,
        CHANNEL_TYPES.guild_thread_private,
    ))


@deprecated_import
@copy_docs(ChannelBase)
class ChannelGuildBase(ChannelBase):
    allowed_types = frozenset((
        CHANNEL_TYPES.guild_text,
        CHANNEL_TYPES.guild_voice,
        CHANNEL_TYPES.guild_category,
        CHANNEL_TYPES.guild_announcements,
        CHANNEL_TYPES.guild_store,
        CHANNEL_TYPES.guild_thread_announcements,
        CHANNEL_TYPES.guild_thread_public,
        CHANNEL_TYPES.guild_thread_private,
        CHANNEL_TYPES.guild_stage,
        CHANNEL_TYPES.guild_directory,
        CHANNEL_TYPES.guild_forum,
    ))


@deprecated_import
@copy_docs(ChannelBase)
class ChannelGuildMainBase(ChannelGuildBase):
    allowed_types = frozenset((
        CHANNEL_TYPES.guild_text,
        CHANNEL_TYPES.guild_voice,
        CHANNEL_TYPES.guild_category,
        CHANNEL_TYPES.guild_announcements,
        CHANNEL_TYPES.guild_store,
        CHANNEL_TYPES.guild_stage,
        CHANNEL_TYPES.guild_directory,
        CHANNEL_TYPES.guild_forum,
    ))


@deprecated_import
@copy_docs(ChannelBase)
class ChannelCategory(ChannelGuildMainBase):
    type = CHANNEL_TYPES.guild_category
    
    allowed_types = frozenset((
        CHANNEL_TYPES.guild_category,
    ))


@deprecated_import
@copy_docs(ChannelBase)
class ChannelDirectory(ChannelGuildMainBase):
    type = CHANNEL_TYPES.guild_directory
    
    allowed_types = frozenset((
        CHANNEL_TYPES.guild_directory,
    ))


@deprecated_import
@copy_docs(ChannelBase)
class ChannelForum(ChannelGuildMainBase):
    type = CHANNEL_TYPES.guild_forum
    
    allowed_types = frozenset((
        CHANNEL_TYPES.guild_forum,
    ))


@deprecated_import
@copy_docs(ChannelBase)
class ChannelStore(ChannelGuildMainBase):
    allowed_types = frozenset((
        CHANNEL_TYPES.guild_store,
    ))


@deprecated_import
@copy_docs(ChannelBase)
class ChannelText(ChannelGuildMainBase, ChannelTextBase):
    type = CHANNEL_TYPES.guild_text
    
    allowed_types = frozenset((
        CHANNEL_TYPES.guild_text,
        CHANNEL_TYPES.guild_announcements,
    ))


@deprecated_import
@copy_docs(ChannelBase)
class ChannelGuildUndefined(ChannelGuildMainBase):
    allowed_types = frozenset((
        7,
        8,
    ))


@deprecated_import
@copy_docs(ChannelBase)
class ChannelVoiceBase(ChannelGuildMainBase):
    allowed_types = frozenset((
        CHANNEL_TYPES.guild_voice,
        CHANNEL_TYPES.guild_stage,
    ))


@deprecated_import
@copy_docs(ChannelBase)
class ChannelVoice(ChannelVoiceBase):
    type = CHANNEL_TYPES.guild_voice
    
    allowed_types = frozenset((
        CHANNEL_TYPES.guild_voice,
    ))


@deprecated_import
@copy_docs(ChannelBase)
class ChannelStage(ChannelVoiceBase):
    type = CHANNEL_TYPES.guild_stage
    
    allowed_types = frozenset((
        CHANNEL_TYPES.guild_stage,
    ))


@deprecated_import
@copy_docs(ChannelBase)
class ChannelPrivate(ChannelTextBase, ChannelBase):
    type = CHANNEL_TYPES.private
    
    allowed_types = frozenset((
        CHANNEL_TYPES.private,
    ))


@deprecated_import
@copy_docs(ChannelBase)
class ChannelGroup(ChannelTextBase, ChannelBase):
    type = CHANNEL_TYPES.private_group
    
    allowed_types = frozenset((
        CHANNEL_TYPES.private_group,
    ))


@deprecated_import
@copy_docs(ChannelBase)
class ChannelThread(ChannelGuildBase, ChannelTextBase):
    allowed_types = frozenset((
        CHANNEL_TYPES.guild_thread_announcements,
        CHANNEL_TYPES.guild_thread_public,
        CHANNEL_TYPES.guild_thread_private,
    ))



__all__ = ('LocationEntityMetadata', 'ScheduledEventEntityMetadata', 'StageEntityMetadata')

from scarletio import RichAttributeErrorBaseType, copy_docs

from ..user import ClientUserBase, create_partial_user_from_id


def try_get_scheduled_event_metadata_type_from_data(data):
    """
    Tries to detect what type of scheduled metadata the given data is.
    
    Parameters
    ----------
    data : `dict` of (`str`, `str`) items
        Scheduled event metadata data.
    
    Returns
    -------
    metadata_type : `None`, `type`
    """
    if 'location' in data:
        metadata_type = LocationEntityMetadata
    elif 'speaker_ids' in data:
        metadata_type = StageEntityMetadata
    else:
        metadata_type = None
    
    return metadata_type


class ScheduledEventEntityMetadata(RichAttributeErrorBaseType):
    """
    Base class for ``ScheduledEvent``'s entity metadata.
    """
    __slots__ = ()
    
    def __new__(cls):
        """
        Creates a new entity metadata instance.
        """
        raise NotImplemented
        
    def __repr__(self):
        """Returns the entity metadata's representation."""
        return f'<{self.__class__.__name__}>'
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a new scheduled event entity metadata instance from the given data.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Entity metadata structure.
        
        Returns
        -------
        self : ``ScheduledEventEntityMetadata``
        """
        raise NotImplemented
    
    
    def to_data(self):
        """
        Converts the entity metadata to json serializable object.
        
        Returns
        -------
        data : `dict` of (`str`, `Any`) items
        """
        return {}
    
    
    def __eq__(self, other):
        """Returns whether the two entity metadatas equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        return True
    
    
    def __hash__(self):
        """Returns the entity metadata's hash value."""
        return 0


class StageEntityMetadata(ScheduledEventEntityMetadata):
    """
    Stage entity metadata of ``ScheduledEvent``-s.
    
    Attributes
    ----------
    speaker_ids : `None`, `tuple` of `int`
        The speakers' identifier of the stage channel.
    """
    __slots__ = ('speaker_ids', )
    
    def __new__(cls, speakers):
        """
        Creates a new stage entity metadata for ``ScheduledEvent``-s.
        
        Parameters
        ----------
        speakers : `None`, `int`, ``ClientUserBase``, `iterable` of (`int`, `ClientUserBase``)
            Speakers of the stage channel.
        
        Raises
        ------
        TypeError
            - If `speakers` type is incorrect.
            - If a speaker's type is incorrect.
        """
        if speakers is None:
            speaker_ids = None
        
        elif isinstance(speakers, ClientUserBase):
            speaker_ids = (ClientUserBase.id, )
        
        elif isinstance(speakers, int):
            speaker_ids = (speakers, )
        
        else:
            iterator = getattr(type(speakers), '__iter__', None)
            if iterator is None:
                raise TypeError(
                    f'`speakers` can be `None`, `int`, `{ClientUserBase.__name__}`, `iterable` of (`int`, '
                    f'`{ClientUserBase.__name__}`), got {speakers.__class__.__name__}; {speakers!r}.'
                )
            
            speaker_ids = None
            
            for speaker in iterator(speakers):
                if isinstance(speaker, ClientUserBase):
                    speaker_id = speaker.id
                
                elif isinstance(speaker, int):
                    speaker_id = speaker
                
                else:
                    raise TypeError(
                        f'`speakers` can contain `int`, `{ClientUserBase.__name__}` elements, got '
                        f'{speaker.__class__.__name__}; {speaker!r}.'
                    )
                
                if speaker_ids is None:
                    speaker_ids = set()
                
                speaker_ids.add(speaker_id)
            
            if (speaker_ids is not None):
                speaker_ids = tuple(sorted(speaker_ids))
            
            else:
                speaker_ids = None
        
        self = object.__new__(cls)
        self.speaker_ids = speaker_ids
        return self
    
    
    @classmethod
    @copy_docs(ScheduledEventEntityMetadata.from_data)
    def from_data(cls, data):
        speaker_ids = data.get('speaker_ids', None)
        if (speaker_ids is None) or (not speaker_ids):
            speaker_ids = None
        else:
            speaker_ids = tuple(sorted(int(speaker_id) for speaker_id in speaker_ids))
        
        self = object.__new__(cls)
        self.speaker_ids = speaker_ids
        return self
    
    
    @classmethod
    @copy_docs(ScheduledEventEntityMetadata.to_data)
    def to_data(self):
        data = {}
        
        speaker_ids = self.speaker_ids
        if (speaker_ids is None):
            speaker_ids = ()
        data['speaker_ids'] = speaker_ids
        
        return data
    
    
    @copy_docs(ScheduledEventEntityMetadata.__eq__)
    def __eq__(self, other):
        if type(self) is not type(other):
            return NotImplemented
        
        if self.speaker_ids != other.speaker_ids:
            return False
        
        return True
    
    
    @copy_docs(ScheduledEventEntityMetadata.__hash__)
    def __hash__(self):
        hash_value = 0
        
        speaker_ids = self.speaker_ids
        if (speaker_ids is not None):
            hash_value ^= len(speaker_ids)
            
            for speaker_id in speaker_ids:
                hash_value ^= speaker_id
        
        return hash_value
    
    
    @property
    def iter_speaker_ids(self):
        """
        Iterates over the speakers' identifiers of the stage channel.
        
        This method is an iterable generator.
        
        Yields
        ------
        speaker_id : `int`
        """
        speaker_ids = self.speaker_ids
        if (speaker_ids is not None):
            yield from speaker_ids
    
    @property
    def speakers(self):
        """
        Returns the speakers of the stage channel.
        
        Returns
        -------
        speakers : `None`, `tuple` of ``ClientUserBase``
        """
        speaker_ids = self.speaker_ids
        if (speaker_ids is not None):
            return tuple(create_partial_user_from_id(speaker_id) for speaker_id in speaker_ids)
    
    
    def iter_speakers(self):
        """
        Iterates over the speakers of the stage channel.
        
        This method is an iterable generator.
        
        Yields
        ------
        speakers : ``ClientUserBase``
        """
        speaker_ids = self.speaker_ids
        if (speaker_ids is not None):
            
            for speaker_id in speaker_ids:
                yield create_partial_user_from_id(speaker_id)


class LocationEntityMetadata(ScheduledEventEntityMetadata):
    """
    Location entity metadata of ``ScheduledEvent``-s.
    
    Attributes
    ----------
    location : `None`, `str`
        The place where the event will take place.
    """
    __slots__ = ('location', )
    
    def __new__(cls, location):
        """
        Creates a new location entity metadata for ``ScheduledEvent``-s.
        
        Parameters
        ----------
        location : `str`
            The location.
        
        Raises
        ------
        TypeError
            If `location`'s type is incorrect.
        ValueError
            If `location` is an empty string.
        """
        if not isinstance(location, str):
            raise TypeError(
                f'`location` can be `str`, got {location.__class__.__name__}; {location!r}.'
            )
        
        if not location:
            raise ValueError(
                f'`location` cannot be empty string.'
            )
        
        self = object.__new__(cls)
        self.location = location
        return self
    
    
    @classmethod
    @copy_docs(ScheduledEventEntityMetadata.from_data)
    def from_data(cls, data):
        location = data.get('location', None)
        
        self = object.__new__(cls)
        self.location = location
        return self
    
    
    @classmethod
    @copy_docs(ScheduledEventEntityMetadata.to_data)
    def to_data(self):
        return {
            'location': self.location,
        }
    
    
    @copy_docs(ScheduledEventEntityMetadata.__eq__)
    def __eq__(self, other):
        if type(self) is not type(other):
            return NotImplemented
        
        if self.location != other.location:
            return False
        
        return True
    
    @copy_docs(ScheduledEventEntityMetadata.__hash__)
    def __hash__(self):
        hash_value = 0
        
        location = self.location
        if (location is not None):
            hash_value ^= hash(location)
        
        return hash_value



__all__ = ('ApplicationCommandAutocompleteInteraction', 'ApplicationCommandAutocompleteInteractionOption')

import reprlib

from scarletio import copy_docs

from ..application_command.preinstanced import ApplicationCommandOptionType

from .interaction_field_base import InteractionFieldBase


class ApplicationCommandAutocompleteInteraction(InteractionFieldBase):
    """
    Represents an ``ApplicationCommand``'s auto completion interaction.
    
    Attributes
    ----------
    id : `int`
        The represented application command's identifier number.
    name : `str`
        The name of the command. It's length can be in range [1:32].
    options : `None`, `tuple` of ``ApplicationCommandAutocompleteOption``
        Parameter auto completion options.
    """
    __slots__ = ('id', 'name', 'options',)
    
    @copy_docs(InteractionFieldBase.__new__)
    def __new__(cls, data, interaction_event):
        # id
        id_ = int(data['id'])
        
        # name
        name = data['name']
        
        # options
        option_datas = data.get('options', None)
        if (option_datas is None) or (not option_datas):
            options = None
        else:
            options = tuple(
                ApplicationCommandAutocompleteInteractionOption(option_data) for option_data in option_datas
            )
        
        self = object.__new__(cls)
        
        self.id = id_
        self.name = name
        self.options = options
        
        return self
    
    
    @copy_docs(InteractionFieldBase.__repr__)
    def __repr__(self):
        repr_parts = [
            '<', self.__class__.__name__,
            ' id=', repr(self.id),
            ', name=', repr(self.name),
        ]
        
        options = self.options
        if (options is not None):
            repr_parts.append(', options=[')
            
            index = 0
            limit = len(options)
            
            while True:
                option = options[index]
                index += 1
                repr_parts.append(repr(option))
                
                if index == limit:
                    break
                
                repr_parts.append(', ')
                continue
            
            repr_parts.append(']')
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    @copy_docs(InteractionFieldBase.__hash__)
    def __hash__(self):
        hash_value = 0
        
        # id
        hash_value ^= self.id
        
        # name
        hash_value ^= hash(self.name)
        
        # options
        options = self.options
        if (options is not None):
            hash_value ^= (len(options) << 8)
            
            for option in options:
                hash_value ^= hash(option)
        
        return hash_value
    
    
    @copy_docs(InteractionFieldBase.__eq__)
    def __eq__(self, other):
        if type(self) is not type(other):
            return NotImplemented
        
        # id
        if self.id != other.id:
            return False
        
        # name
        if self.name != other.name:
            return False
        
        # options
        if self.options != other.options:
            return False
        
        return True
        
    
    @property
    def focused_option(self):
        """
        Returns the focused option of the application command autocomplete interaction.
        
        Returns
        -------
        option : `None`, ``ApplicationCommandAutocompleteInteractionOption``
        """
        options = self.options
        if options is None:
            focused_option = None
        else:
            for option in options:
                focused_option = option.focused_option
                if (focused_option is not None):
                    break
            else:
                focused_option = None
        
        return focused_option
    
    
    def get_non_focused_values(self):
        """
        Gets the non focused values of the interaction.
        
        Returns
        -------
        non_focused_options : `dict` of (`str`, (`None`, `str`)) items
        """
        return dict(self._iter_non_focused_values())
    
    
    def _iter_non_focused_values(self):
        """
        Iterates over the non focused values of the interaction.
        
        This method is an iterable generator.
        
        Yields
        ------
        name : `str`
            The option's name.
        value : `None`, `str`
            The option's value.
        """
        options = self.options
        if (options is not None):
            for option in options:
                yield from option._iter_non_focused_values()
    
    
    def get_value_of(self, *option_names):
        """
        Gets the value for the option by the given name.
        
        Parameters
        ----------
        *option_names : `str`
            The option(s)'s name.
        
        Returns
        -------
        value : `None`, `str`
            The value, the user has been typed.
        """
        if option_names:
            option_name, *option_names = option_names
            
            options = self.options
            if options is None:
                value = None
            else:
                for option in options:
                    if option.name == option_name:
                        value = option.get_value_of(*option_names)
                        break
                else:
                    value = None
        else:
            value = None
        
        return value
    
    
    @property
    def value(self):
        """
        Returns the focused option's value of the application command autocomplete interaction.
        
        Returns
        -------
        value : `None`, `str`
        """
        focused_option = self.focused_option
        if (focused_option is None):
            value = None
        else:
            value = focused_option.value
        
        return value



class ApplicationCommandAutocompleteInteractionOption:
    """
    Application auto complete option representing an auto completable parameters.
    
    Attributes
    ----------
    focused : `bool`
        Whether this field is focused by the user.
    name : `str`
        The name of the parameter.
    options : `None`, `tuple` of ``ApplicationCommandAutocompleteInteractionOption``
        Nested functions.
    type : ``ApplicationCommandOptionType``
        The represented option's type.
    value : `None`, `str`
        The value, the user has been typed.
    """
    __slots__ = ('focused', 'name', 'options', 'type', 'value')
    
    def __new__(cls, data):
        """
        Creates a new ``ApplicationCommandAutocompleteOption`` from the data received from Discord.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`)
            Application command autocomplete option data.
        """
        # focused
        focused = data.get('focused', False)
        
        # name
        name = data['name']
        
        # options
        option_datas = data.get('options', None)
        if (option_datas is None) or (not option_datas):
            options = None
        else:
            options = tuple(
                ApplicationCommandAutocompleteInteractionOption(option_data) for option_data in option_datas
            )
        
        # type
        type_ = ApplicationCommandOptionType.get(data['type'])
        
        # value
        value = data.get('value', None)
        if (value is not None) and (not value):
            value = None
        
        self = object.__new__(cls)
        self.focused = focused
        self.name = name
        self.options = options
        self.type = type_
        self.value = value
        return self
    
    
    def __repr__(self):
        """Returns the application command autocomplete option's representation."""
        repr_parts = [
            '<', self.__class__.__name__,
            ' name=', repr(self.name),
        ]
        
        if self.focused:
            repr_parts.append(' (focused)')
        
        value = self.value
        if (value is not None):
            repr_parts.append(', value=')
            repr_parts.append(reprlib.repr(value))
        
        type_ = self.type
        repr_parts.append(', type=')
        repr_parts.append(type_.name)
        repr_parts.append(' (')
        repr_parts.append(repr(type_.value))
        repr_parts.append(')')
        
        options = self.options
        if (options is not None):
            repr_parts.append(', options=[')
            
            index = 0
            limit = len(options)
            
            while True:
                option = options[index]
                index += 1
                repr_parts.append(repr(option))
                
                if index == limit:
                    break
                
                repr_parts.append(', ')
                continue
            
            repr_parts.append(']')
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __hash__(self):
        """Returns the application command autocomplete option's representation."""
        hash_value = 0
        
        # focused
        hash_value ^= (self.focused << 16)
        
        # name
        hash_value ^= hash(self.name)
        
        # options
        options = self.options
        if (options is not None):
            hash_value ^= (len(options) << 8)
            
            for option in options:
                hash_value ^= hash(option)
        
        # type
        hash_value ^= self.type.value
        
        # value
        value = self.value
        if (value is not None):
            hash_value ^= hash(value)
        
        return hash_value
    
    
    def __eq__(self, other):
        """Returns whether the two application command autocomplete option's are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        # focused
        if self.focused != other.focused:
            return False
        
        # name
        if self.name != other.name:
            return False
        
        # options
        if self.options != other.options:
            return False
        
        # type
        if self.type is not other.type:
            return False
        
        # value
        if self.value != other.value:
            return False
        
        return True
    
    
    @property
    def focused_option(self):
        """
        Returns the focused option of the application command autocomplete interaction option.
        
        Returns
        -------
        option : `None`, ``ApplicationCommandAutocompleteInteractionOption``
        """
        if self.focused:
            return self
        
        options = self.options
        if (options is not None):
            for option in options:
                focused_option =  option.focused_option
                if (focused_option is not None):
                    return focused_option
    
    
    def _iter_non_focused_values(self):
        """
        Iterates over the non focused values of the interaction option.
        
        This method is an iterable generator.
        
        Yields
        ------
        name : `str`
            The option's name.
        value : `None`, `str`
            The option's value.
        """
        type_ = self.type
        if (
            (type_ is ApplicationCommandOptionType.sub_command) or
            (type_ is ApplicationCommandOptionType.sub_command_group)
        ):
            options = self.options
            if (options is not None):
                for option in options:
                    yield from option._iter_non_focused_values()
        else:
            if not self.focused:
                yield self.name, self.value
    
    
    def get_value_of(self, *option_names):
        """
        Gets the value for the option by the given name.
        
        Parameters
        ----------
        *option_names : `str`
            The option(s)'s name.
        
        Returns
        -------
        value : `None`, `str`
            The value, the user has been typed.
        """
        if option_names:
            option_name, *option_names = option_names
            
            options = self.options
            if options is None:
                value = None
            else:
                for option in options:
                    if option.name == option_name:
                        value = option.get_value_of(*option_names)
                        break
                else:
                    value = None
        else:
            value = self.value
        
        return value
